# -*- coding: utf-8 -*-

"""
Test Protocols in the CLI app.
"""

from __future__ import absolute_import, unicode_literals
import logging

import pytest

from tests.fixtures import (attribute, attributes, client, config, device,
                            interface, network, protocol_type, site,
                            site_client)
from tests.fixtures.circuits import (circuit, circuit_attributes, interface_a,
                                     interface_z, device_a, device_z)
from tests.fixtures.protocols import (protocol, protocols, protocol_attribute,
                                      protocol_attribute2)

from tests.util import CliRunner, assert_output

log = logging.getLogger(__name__)


def test_protocols_add(site_client, device_a, interface_a, site, protocol_type):
    """Test ``nsot protocol add``."""

    device_id = str(device_a['id'])
    interface_id = str(interface_a['id'])

    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        # Add a protocol.
        result = runner.run(
            "protocols add -t bgp -D %s -I %s -e 'my new proto'" % (device_id, interface_id)
        )
        assert result.exit_code == 0
        assert 'Added protocol!' in result.output

        # Verify addition.
        result = runner.run('protocols list')
        assert result.exit_code == 0
        assert 'bgp' in result.output
        assert device_a['hostname'] in result.output
        assert 'my new proto' in result.output

        # Add a second protocol with attributes.
        attributes = 'foo=test_attribute'
        result = runner.run(
            "protocols add -t bgp -D %s -I %s -a %s" % (device_id, interface_id, attributes)
        )
        assert result.exit_code == 0
        assert 'Added protocol!' in result.output

        # Verify addition.
        result = runner.run('protocols list')
        assert result.exit_code == 0
        assert 'bgp' in result.output
        assert device_a['hostname'] in result.output
        assert attributes in result.output


def test_protocols_list(site_client, device_a, interface_a, site, circuit, protocol):
    """Test ``nsot protocols list``"""

    device_id = str(device_a['id'])
    interface_id = str(interface_a['id'])
    protocol_id = str(protocol['id'])

    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        result = runner.run('protocols list')
        assert result.exit_code == 0
        assert 'bgp' in result.output

        # test -t/--type
        result = runner.run('protocols list -t bgp')
        assert result.exit_code == 0
        assert 'bgp' in result.output

        # Test -D/--device
        result = runner.run('protocols list -D %s' % device_id)
        assert result.exit_code == 0
        assert device_a['hostname'] in result.output

        # Test -I/--interface
        result = runner.run('protocols list -I %s' % interface_id)
        assert result.exit_code == 0
        assert interface_a['name_slug'] in result.output

        # Test -a/--attributes
        result = runner.run('protocols list -a foo=test_protocol')
        assert result.exit_code == 0
        assert protocol['attributes']['foo'] in result.output

        # Test -c/--circuit
        result = runner.run('protocols list -c %s' % circuit['name'])
        assert result.exit_code == 0
        assert circuit['name'] in result.output

        # Test -e/--description
        result = runner.run('protocols list -e "%s"' % protocol['description'])
        assert result.exit_code == 0
        assert protocol['description'] in result.output

        # Test -I/--id
        result = runner.run('protocols list -i %s' % protocol_id)
        assert result.exit_code == 0
        assert protocol_id in result.output

        # Test -q/--query
        slug = '{device}:{type}:{id}'.format(**protocol)
        result = runner.run('protocols list -q foo=test_protocol')
        assert result.exit_code == 0
        assert slug in result.output


def test_protocols_list_limit(site_client, protocols):
    """
    If ``--limit 2`` is used, we should only see the first two Protocol objects
    """
    limit = 2
    runner = CliRunner(site_client.config)

    with runner.isolated_filesystem():
        result = runner.run('protocols list -l {}'.format(limit))
        assert result.exit_code == 0

        expected_protocols = protocols[:limit]
        unexpected_protocols = protocols[limit:]

        for p in expected_protocols:
            assert p['device'] in result.output
        for p in unexpected_protocols:
            assert p['device'] not in result.output


def test_protocols_list_offset(site_client, protocols):
    """
    If ``--limit 2`` and ``--offset 2`` are used, we should only see the third
    and fourth Protocol objects that were created
    """
    limit = 2
    offset = 2
    runner = CliRunner(site_client.config)

    with runner.isolated_filesystem():
        result = runner.run('protocols list -l {} -o {}'.format(limit, offset))
        assert result.exit_code == 0

        expected_protocols = protocols[offset:limit+offset]
        unexpected_protocols = protocols[limit+offset:]

        for p in expected_protocols:
            assert p['device'] in result.output
        for p in unexpected_protocols:
            assert p['device'] not in result.output


def test_protocols_update(site_client, interface_a, device_a, site, circuit, protocol, protocol_attribute):
    site_id = str(protocol['site'])
    proto_id = protocol['id']
    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        # Update description
        result = runner.run('protocols update -i %s -e "bees buzz"' % proto_id)
        assert result.exit_code == 0
        assert 'Updated protocol!' in result.output

        # Ensure that buzz is not the bizness
        result = runner.run('protocols list')
        assert result.exit_code == 0
        assert 'buzz' in result.output
        assert 'bizness' not in result.output

        # Add an attribute
        result = runner.run(
            'protocols update -i %s --add-attributes -a boo=test_attribute' % proto_id
        )
        assert result.exit_code == 0
        assert 'Updated protocol!' in result.output

        result = runner.run('protocols list')
        assert result.exit_code == 0
        assert 'test_attribute' in result.output

        # Add an attribute without using --add-attributes.
        result = runner.run(
            'protocols update -i %s -a boo=test_attribute' % proto_id
        )
        assert result.exit_code == 0
        assert 'Updated protocol!' in result.output

        result = runner.run('protocols list')
        assert result.exit_code == 0
        assert 'test_attribute' in result.output

        # Delete an attribute
        result = runner.run(
            'protocols update -i %s --delete-attributes -a foo=test_protocol' % proto_id
        )
        assert result.exit_code == 0
        assert 'Updated protocol!' in result.output

        result = runner.run('protocols list')
        assert result.exit_code == 0
        assert 'test_protocol' not in result.output

        # Replace an attribute
        result = runner.run(
            'protocols update -i %s --replace-attributes -a foo=test_replace' % proto_id
        )
        assert result.exit_code == 0
        assert 'Updated protocol!' in result.output

        result = runner.run('protocols list')
        assert result.exit_code == 0
        assert 'test_protocol' not in result.output
        assert 'test_replace' in result.output


def test_protocols_remove(site_client, protocol):
    runner = CliRunner(site_client.config)

    with runner.isolated_filesystem():
        result = runner.run('protocols remove -i %s -s %s' % (protocol['id'], protocol['site']))
        assert result.exit_code == 0

        result = runner.run('protocols list')
        assert result.exit_code == 0
        assert 'No protocol found' in result.output



# -*- coding: utf-8 -*-

"""
Sub-command for Protocol Types.

In all cases ``data = ctx.params`` when calling the appropriate action method
on ``ctx.obj``. (e.g. ``ctx.obj.add(ctx.params)``)

Also, ``action = ctx.info_name`` *might* reliably contain the name of the
action function, but still not sure about that. If so, every function could be
fundamentally simplified to this::

    getattr(ctx.obj, ctx.info_name)(ctx.params)
"""

from __future__ import unicode_literals
import logging

from ..vendor import click
from . import callbacks, types

# Logger
log = logging.getLogger(__name__)

# Ordered list of 2-tuples of (field, display_name) used to translate object
# field names oto their human-readable form when calling .print_list().
DISPLAY_FIELDS = (
    ('id', 'ID'),
    ('name', 'Name'),
    ('description', 'Description'),
    ('required_attributes', 'Required Attributes'),
)

# Fields to display when viewing a single record.
VERBOSE_FIELDS = (
    ('id', 'ID'),
    ('name', 'Name'),
    ('description', 'Description'),
    ('required_attributes', 'Required Attributes'),
    ('site', 'Site ID'),
)


# Main group
@click.group()
@click.pass_context
def cli(ctx):
    """
    Protocol Type objects.

    A Protocol Type resource can represent a network protocol type (e.g. bgp,
    is-is, ospf, etc.)

    Protocol Types can have any number of required attributes as defined below.
    """


# Add
@cli.command()
@click.option(
    '-r',
    '--required-attributes',
    metavar='ATTRIBUTE',
    type=str,
    help=('''The name of a Protocol attribute. This option can be providedi
           multiple times, once per attribute.'''),
    multiple=True,
)
@click.option(
    '-d',
    '--description',
    metavar='DESCRIPTION',
    type=str,
    help='The description for this Protocol Type.',
)
@click.option(
    '-n',
    '--name',
    metavar='NAME',
    type=str,
    help='The name of the Protocol Type.',
    required=True,
)
@click.option(
    '-s',
    '--site-id',
    metavar='SITE_ID',
    type=int,
    help='Unique ID of the Site this Protocol Type is under.  [required]',
    callback=callbacks.process_site_id,
)
@click.pass_context
def add(ctx, required_attributes, description, name, site_id):
    """
    Add a new Protocol Type.

    You must provide a Protocol Type name or ID using the UPDATE option.

    When adding a new Protocol Type, you must provide a value for the -n/--name
    option.

    Examples: OSPF, BGP, etc.

    You may also provide required Protocol attributes, you may specify the
    -r/--required-attributes option once for each attribute. The Protocol
    attributes must exist before adding them to a protocol type.

    You must provide a Site ID using the -s/--site-id option.
    """
    data = ctx.params

    # Remove if empty; allow default assignment
    if description is None:
        data.pop('description')

    ctx.obj.add(data)


# List
@cli.group(invoke_without_command=True)
@click.option(
    '-d',
    '--description',
    metavar='DESCRIPTION',
    type=str,
    help='Filter by Protocol Type matching this description.',
)
@click.option(
    '-i',
    '--id',
    metavar='ID',
    type=types.NATURAL_KEY,
    help='Unique ID of the Protocol Type being retrieved.',
)
@click.option(
    '-l',
    '--limit',
    metavar='LIMIT',
    type=int,
    help='Limit result to N resources.',
)
@click.option(
    '-n',
    '--name',
    metavar='NAME',
    help='Filter to Protocol Type matching this name.'
)
@click.option(
    '-o',
    '--offset',
    metavar='OFFSET',
    help='Skip the first N resources.',
)
@click.option(
    '-s',
    '--site-id',
    metavar='SITE_ID',
    help='Unique ID of the Site this Protocol Type is under.',
    callback=callbacks.process_site_id,
)
@click.pass_context
def list(ctx, description, id, limit, name, offset, site_id):
    """
    List existing Protocol Types for a Site.

    You must provide a Site ID using the -s/--site-id option.

    When listing Protocol Types, all objects are displayed by default. You
    optionally may lookup a single Protocol Types by ID using the -i/--id
    option. The ID can either be the numeric ID of the Protocol Type.
    """
    data = ctx.params

    # If we provide ID, show more fields
    if any([id, name]):
        display_fields = VERBOSE_FIELDS
    else:
        display_fields = DISPLAY_FIELDS

    # If we aren't passing a sub-command, just call list(), otherwise let it
    # fallback to default behavior.
    if ctx.invoked_subcommand is None:
        ctx.obj.list(
            data, display_fields=display_fields,
            verbose_fields=VERBOSE_FIELDS
        )


# Remove
@cli.command()
@click.option(
    '-i',
    '--id',
    metavar='ID',
    type=types.NATURAL_KEY,
    help='Unique ID of the Protocol Type being deleted.',
    required=True,
)
@click.option(
    '-s',
    '--site-id',
    metavar='SITE_ID',
    type=int,
    help='Unique ID of the Site this Protocol Type is under.',
    callback=callbacks.process_site_id,
)
@click.pass_context
def remove(ctx, id, site_id):
    """
    Remove an Protocol Type.

    You must provide a Site ID using the -s/--site-id option.

    When removing an Protocol Type, you must provide the ID of the Protocol
    Type using -i/--id.

    You may retrieve the ID for an Protocol Type by parsing it from the list of
    Protocol Types for a given Site:

      nsot protocol_types list --site-id <site_id> | grep <protocol_type name>
    """
    data = ctx.params
    ctx.obj.remove(**data)


# Update
@cli.command()
@click.option(
    '-r',
    '--required-attributes',
    metavar='ATTRIBUTE',
    type=str,
    help=('''The name of a Protocol attribute. This option can be provided multiple
         times, once per attribute.'''),
    multiple=True,
    callback=callbacks.transform_attributes,
)
@click.option(
    '-d',
    '--description',
    metavar='DESCRIPTION',
    type=str,
    help='The description for this Protocol Type.',
)
@click.option(
    '-i',
    '--id',
    metavar='ID',
    type=types.NATURAL_KEY,
    help='Unique ID of the Protocol Type being updated.',
    required=True,
)
@click.option(
    '-n',
    '--name',
    metavar='NAME',
    type=str,
    help='The name of the Protocol Type.',
)
@click.option(
    '-s',
    '--site-id',
    metavar='SITE_ID',
    type=int,
    help='Unique ID of the Site this Protocol Type is under.',
    callback=callbacks.process_site_id,
)
@click.pass_context
def update(ctx, required_attributes, description, id, name, site_id):
    """
    Update an Protocol Type.

    You must provide a Site ID using the -s/--site-id option.

    When updating an Protocol Type you must provide the ID (-i/--id). The ID
    can either be the numeric ID of the Protocol Type or the name.

    The -a/--attributes option may be provided multiple times, once for each
    key-value pair.

    When modifying attributes you have three actions to choose from:

    * Add (--add-attributes). This is the default behavior that will add
    attributes if they don't exist, or update them if they do.

    * Delete (--delete-attributes). This will cause attributes to be
    deleted. If combined with --multi the attribute will be deleted if
    either no value is provided, or if the attribute no longer contains a
    valid value.

    * Replace (--replace-attributes). This will cause attributes to
    replaced. If combined with --multi and multiple attributes of the same
    name are provided, only the last value provided will be used.
    """
    data = ctx.params
    ctx.obj.update(data)



# -*- coding: utf-8 -*-

"""
Sub-command for Protocols.

In all cases ``data = ctx.params`` when calling the appropriate action method
on ``ctx.obj``. (e.g. ``ctx.obj.add(ctx.params)``)

Also, ``action = ctx.info_name`` *might* reliably contain the name of the
action function, but still not sure about that. If so, every function could be
fundamentally simplified to this::

    getattr(ctx.obj, ctx.info_name)(ctx.params)
"""

from __future__ import unicode_literals
import logging

from ..vendor import click
from . import callbacks, types


# Logger
log = logging.getLogger(__name__)

# Ordered list of 2-tuples of (field, display_name) used to translate object
# field names oto their human-readable form when calling .print_list().
DISPLAY_FIELDS = (
    ('id', 'ID'),
    ('device', 'Device'),
    ('type', 'Type'),
    ('interface', 'Interface'),
    ('circuit', 'Circuit'),
    ('attributes', 'Attributes'),
)

# Fields to display when viewing a single record.
VERBOSE_FIELDS = (
    ('id', 'ID'),
    ('device', 'Device'),
    ('type', 'Type'),
    ('interface', 'Interface'),
    ('circuit', 'Circuit'),
    ('auth_string', 'Auth_String'),
    ('description', 'Description'),
    ('site', 'Site'),
    ('attributes', 'Attributes'),
)


# Main group
@click.group()
@click.pass_context
def cli(ctx):
    """
    Protocol objects.

    A Protocol resource can represent a network routing protocol.

    """


# Add
@cli.command()
@click.option(
    '-u',
    '--auth_string',
    metavar='AUTH_STRING',
    default='',
    help='The authentication string (such as MD5 sum).',
)
@click.option(
    '-a',
    '--attributes',
    metavar='ATTRS',
    help='A key/value pair attached to this Protocol (format: key=value).',
    multiple=True,
    callback=callbacks.transform_attributes,
)
@click.option(
    '-c',
    '--circuit',
    metavar='CIRCUIT',
    help='The circuit that this protocol is running over.',
)
@click.option(
    '-D',
    '--device',
    metavar='DEVICE',
    type=types.NATURAL_KEY,
    help=(
        'Unique ID of the Device to which this Protocol is '
        'running on.'
    ),
    required=True,
)
@click.option(
    '-e',
    '--description',
    metavar='DESCRIPTION',
    type=str,
    help='The description for this Protocol.',
)
@click.option(
    '-I',
    '--interface',
    metavar='INTERFACE',
    type=str,
    help=(
        'The Interface this Protocol is running on. Either interface'
        'or circuit must be populated.'
    ),
)
@click.option(
    '-s',
    '--site-id',
    metavar='SITE_ID',
    type=int,
    help=(
        'Unique ID of the Site this Protocol is under.'
        'If not set, this will be inherited off of the device\'s site'
    ),
    callback=callbacks.process_site_id,
)
@click.option(
    '-t',
    '--type',
    metavar='TYPE',
    type=types.NATURAL_KEY,
    help='The type of the protocol.',
    required=True,
)
@click.pass_context
def add(ctx, auth_string, attributes, circuit, device, description, interface,
        site_id, type):
    """
    Add a new Protocol.

    You must provide a Device hostname or ID using the -D/--device option.

    You must provide the type of the protocol (e.g. OSPF, BGP, etc.)

    If you wish to add attributes, you may specify the -a/--attributes
    option once for each key/value pair.

    """
    data = ctx.params

    if interface is None and circuit is None:
        raise click.UsageError(
            '''Must have interface "-i" / "--interface" or
               circuit "-c" / "--circuit" populated'''
        )

    if description is None:
        data.pop('description')

    ctx.obj.add(data)


# List
@cli.group(invoke_without_command=True)
@click.option(
    '-a',
    '--attributes',
    metavar='ATTRS',
    help='A key/value pair attached to this Protocol (format: key=value).',
    multiple=True,
)
@click.option(
    '-u',
    '--auth_string',
    metavar='AUTH_STRING',
    default='',
    help='The authentication string (such as MD5 sum).',
)
@click.option(
    '-c',
    '--circuit',
    metavar='CIRCUIT',
    help='The circuit that this protocol is running over.',
)
@click.option(
    '-d',
    '--delimited',
    is_flag=True,
    help='Display set query results separated by commas vs. newlines.',
    default=False,
    show_default=True,
)
@click.option(
    '-e',
    '--description',
    metavar='DESCRIPTION',
    type=str,
    help='Filter by Protocols matching this description.',
)
@click.option(
    '-D',
    '--device',
    metavar='DEVICE',
    type=types.NATURAL_KEY,
    help=(
        'Unique ID of the Device to which this Protocol is '
        'running on.'
    ),
)
@click.option(
    '-g',
    '--grep',
    is_flag=True,
    help='Display list results in a grep-friendly format.',
    default=False,
    show_default=True,
)
@click.option(
    '-i',
    '--id',
    metavar='ID',
    help='Unique ID of the Protocol being retrieved.',
)
@click.option(
    '-I',
    '--interface',
    metavar='INTERFACE',
    type=types.NATURAL_KEY,
    help=(
        'The Interface this Protocol is running on. Either interface'
        'or circuit must be populated.'
    ),
)
@click.option(
    '-l',
    '--limit',
    metavar='LIMIT',
    type=int,
    help='Limit result to N resources.',
)
@click.option(
    '-o',
    '--offset',
    metavar='OFFSET',
    help='Skip the first N resources.',
)
@click.option(
    '-q',
    '--query',
    metavar='QUERY',
    help='Perform a set query using Attributes and output matching Protocols.'
)
@click.option(
    '-s',
    '--site-id',
    metavar='SITE_ID',
    help='Unique ID of the Site this Protocol is under.',
    callback=callbacks.process_site_id,
)
@click.option(
    '-t',
    '--type',
    metavar='TYPE',
    type=types.NATURAL_KEY,
    help='The protocol type',
)
@click.pass_context
def list(ctx, attributes, auth_string, circuit, delimited, description, device,
         grep, id, interface, limit, offset, query, site_id, type):
    """
    List existing Protocols for a Site.

    You must provide a Site ID using the -s/--site-id option.

    You must provide the protocol type.

    When listing Protocols, all objects are displayed by default. You
    optionally may lookup a single Protocols by ID using the -i/--id option.

    You may look up the protocols for a single Device using -D/--device which
    can be either a hostname or ID for a Device.
    """
    data = ctx.params
    data.pop('delimited')  # We don't want this going to the server.

    # If we provide ID, show more fields
    if id is not None:
        display_fields = VERBOSE_FIELDS
    else:
        display_fields = DISPLAY_FIELDS

    # If we aren't passing a sub-command, just call list(), otherwise let it
    # fallback to default behavior.
    if ctx.invoked_subcommand is None:
        if query is not None:
            ctx.obj.natural_keys_by_query(data, delimited)
        else:
            ctx.obj.list(
                data, display_fields=display_fields,
                verbose_fields=VERBOSE_FIELDS
            )


# Remove
@cli.command()
@click.option(
    '-i',
    '--id',
    metavar='ID',
    help='Unique ID of the Protocol being deleted.',
    required=True,
)
@click.option(
    '-s',
    '--site-id',
    metavar='SITE_ID',
    type=int,
    help='Unique ID of the Site this Protocol is under.',
    callback=callbacks.process_site_id,
    required=True,
)
@click.pass_context
def remove(ctx, id, site_id):
    """
    Remove a Protocol.

    You must provide a Site ID using the -s/--site-id option.

    When removing an Protocol, you must provide the ID of the Protocol using
    -i/--id.

    You may retrieve the ID for an Protocol by parsing it from the list of
    Protocols for a given Site:

        nsot interfaces list --site-id <site_id> | grep <protocol>
    """
    data = ctx.params
    ctx.obj.remove(**data)


# Update
@cli.command()
@click.option(
    '-a',
    '--attributes',
    metavar='ATTRS',
    help='A key/value pair attached to this Protocol (format: key=value).',
    multiple=True,
    callback=callbacks.transform_attributes,
)
@click.option(
    '-u',
    '--auth_string',
    metavar='AUTH_STRING',
    default='',
    help='The authentication string (such as MD5 sum).',
)
@click.option(
    '-c',
    '--circuit',
    metavar='CIRCUIT',
    help='The circuit that this protocol is running over.',
)
@click.option(
    '-e',
    '--description',
    metavar='DESCRIPTION',
    type=str,
    help='The description for this Protocol.',
)
@click.option(
    '-D',
    '--device',
    metavar='DEVICE',
    type=types.NATURAL_KEY,
    help=(
        'Unique ID of the Device to which this Protocol is '
        'running on.'
    ),
)
@click.option(
    '-i',
    '--id',
    metavar='ID',
    type=int,
    help='Unique ID or natural key of the Protocol being updated.',
    required=True,
)
@click.option(
    '-I',
    '--interface',
    metavar='INTERFACE',
    type=types.NATURAL_KEY,
    help=(
        'The Interface this Protocol is running on. Either interface'
        'or circuit must be populated.'
    ),
)
@click.option(
    '-s',
    '--site-id',
    metavar='SITE_ID',
    type=int,
    help='Unique ID of the Site this Protocol is under.',
    callback=callbacks.process_site_id,
)
@click.option(
    '-t',
    '--type',
    metavar='TYPE',
    type=types.NATURAL_KEY,
    help='The Protocol type ID.',
)
@click.option(
    '--add-attributes',
    'attr_action',
    flag_value='add',
    default=True,
    help=(
        'Causes attributes to be added. This is the default and providing it '
        'will have no effect.'
    )
)
@click.option(
    '--delete-attributes',
    'attr_action',
    flag_value='delete',
    help=(
        'Causes attributes to be deleted instead of updated. If combined with'
        'with --multi the attribute will be deleted if either no value is '
        'provided, or if attribute no longer has an valid values.'
    ),
)
@click.option(
    '--replace-attributes',
    'attr_action',
    flag_value='replace',
    help=(
        'Causes attributes to be replaced instead of updated. If combined '
        'with --multi, the entire list will be replaced.'
    ),
)
@click.option(
    '--multi',
    is_flag=True,
    help='Treat the specified attributes as a list type.',
)
@click.pass_context
def update(ctx, attributes, auth_string, circuit, description, device, id,
           interface, site_id, type, attr_action, multi):
    """
    Update a Protocol.

    You must provide a Site ID using the -s/--site-id option.

    You must provide the type of the protocol (e.g. OSPF, BGP, etc.)

    When updating an Protocol you must provide the ID (-i/--id) and at least
    one of the optional arguments.

    The -a/--attributes option may be provided multiple times, once for each
    key-value pair.

    When modifying attributes you have three actions to choose from:

    * Add (--add-attributes). This is the default behavior that will add
    attributes if they don't exist, or update them if they do.

    * Delete (--delete-attributes). This will cause attributes to be
    deleted. If combined with --multi the attribute will be deleted if
    either no value is provided, or if the attribute no longer contains a
    valid value.

    * Replace (--replace-attributes). This will cause attributes to
    replaced. If combined with --multi and multiple attributes of the same
    name are provided, only the last value provided will be used.
    """
    if not any([attributes, description, type, auth_string, circuit,
                interface]):
        msg = 'You must supply at least one of the optional arguments.'
        raise click.UsageError(msg)
    ctx.obj.update(ctx.params)



# -*- coding: utf-8 -*-

"""
Sub-command for Values.

In all cases ``data = ctx.params`` when calling the appropriate action method
on ``ctx.obj``. (e.g. ``ctx.obj.add(ctx.params)``)

Also, ``action = ctx.info_name`` *might* reliably contain the name of the
action function, but still not sure about that. If so, every function could be
fundamentally simplified to this::

    getattr(ctx.obj, ctx.info_name)(ctx.params)
"""

from __future__ import unicode_literals

from ..vendor import click
from . import callbacks


__author__ = 'Jathan McCollum'
__maintainer__ = 'Jathan McCollum'
__email__ = 'jathan@dropbox.com'
__copyright__ = 'Copyright (c) 2016 Dropbox, Inc.'


# Ordered list of 2-tuples of (field, display_name) used to translate object
# field names oto their human-readable form when calling .print_list().
DISPLAY_FIELDS = (
    ('name', 'Name'),
    ('value', 'Value'),
    ('resource_name', 'Resource'),
    ('resource_id', 'Resource ID'),
    # ('site_id': 'Site ID'),
)

# Fields to display when viewing a single record.
VERBOSE_FIELDS = (
    ('id', 'ID'),
    ('name', 'Name'),
    ('value', 'Value'),
    ('resource_name', 'Resource'),
    ('resource_id', 'Resource ID'),
    ('attribute', 'Attribute'),
)


# Main group
@click.group()
@click.pass_context
def cli(ctx):
    """
    Value objects.

    Values are assigned by Attribute to various resource objects.
    """


# List
@cli.command()
@click.option(
    '-n',
    '--name',
    metavar='NAME',
    required=True,
    help='Filter to Attribute with this name.',
)
@click.option(
    '-r',
    '--resource-name',
    metavar='RESOURCE_NAME',
    help='Filter to Values for a specific resource name (e.g. Device)',
    callback=callbacks.transform_resource_name,
)
@click.option(
    '-s',
    '--site-id',
    metavar='SITE_ID',
    type=int,
    help='Unique ID of the Site this Attribute is under.  [required]',
    callback=callbacks.process_site_id,
)
@click.pass_context
def list(ctx, name, resource_name, site_id):
    """
    List existing Values by Attribute name for a Site.

    You must provide a Name using the -n/--name option.

    You must provide a Site ID using the -s/--site-id option.

    When listing Values, all matching Attributes across all resource types are
    returned by default. You may optionally filter by a single resource name by
    using the -r/--resource-name option.
    """
    data = ctx.params

    # Fetch the matching values directly from the API client, filter
    # duplicates, and sort the output.
    results = ctx.obj.api.sites(site_id).values.get(**data)
    values = {d['value'] for d in results}
    click.echo('\n'.join(sorted(values)))



# -*- coding: utf-8 -*-

"""
Sub-command for Interfaces.

In all cases ``data = ctx.params`` when calling the appropriate action method
on ``ctx.obj``. (e.g. ``ctx.obj.add(ctx.params)``)

Also, ``action = ctx.info_name`` *might* reliably contain the name of the
action function, but still not sure about that. If so, every function could be
fundamentally simplified to this::

    getattr(ctx.obj, ctx.info_name)(ctx.params)
"""

from __future__ import unicode_literals
import logging

from ..vendor import click
from . import callbacks, types
from .cmd_networks import DISPLAY_FIELDS as NETWORK_DISPLAY_FIELDS


# Logger
log = logging.getLogger(__name__)

# Ordered list of 2-tuples of (field, display_name) used to translate object
# field names oto their human-readable form when calling .print_list().
DISPLAY_FIELDS = (
    ('id', 'ID'),
    ('name_slug', 'Name (Key)'),
    ('parent', 'Parent'),
    ('mac_address', 'MAC'),
    ('addresses', 'Addresses'),
    ('attributes', 'Attributes'),
)

# Fields to display when viewing a single record.
VERBOSE_FIELDS = (
    ('id', 'ID'),
    ('name_slug', 'Name (Key)'),
    ('parent', 'Parent'),
    ('mac_address', 'MAC'),
    ('addresses', 'Addresses'),
    ('speed', 'Speed'),
    ('type', 'Type'),
    ('attributes', 'Attributes'),
)


# Main group
@click.group()
@click.pass_context
def cli(ctx):
    """
    Interface objects.

    An Interface resource can represent a network interface attached to a
    Device. Working with interfaces is usually the device and an interface
    name.

    Interfaces can have any number of arbitrary attributes as defined below.
    """


# Add
@cli.command()
@click.option(
    '-a',
    '--attributes',
    metavar='ATTRS',
    help='A key/value pair attached to this Interface (format: key=value).',
    multiple=True,
    callback=callbacks.transform_attributes,
)
@click.option(
    '-c',
    '--addresses',
    metavar='ADDRS',
    help='An IPv4/IPv6 address in CIDR format.',
    multiple=True,
)
@click.option(
    '-D',
    '--device',
    metavar='DEVICE',
    type=types.NATURAL_KEY,
    help=(
        'Unique ID of the Device to which this Interface is '
        'attached.  [required]'
    )
)
@click.option(
    '-e',
    '--description',
    metavar='DESCRIPTION',
    type=str,
    help='The description for this Interface.',
)
@click.option(
    '-m',
    '--mac-address',
    metavar='MAC',
    type=str,
    help='The MAC address of the Interface.',
)
@click.option(
    '-n',
    '--name',
    metavar='NAME',
    type=str,
    help='The name of the Interface.  [required]',
)
@click.option(
    '-p',
    '--parent-id',
    metavar='PARENT_ID',
    type=types.NATURAL_KEY,
    help='Unique ID of the parent interface.'
)
@click.option(
    '-S',
    '--speed',
    metavar='SPEED',
    type=int,
    help='The Interface speed in Mbps (e.g. 20000 for 20Gps).',
)
@click.option(
    '-s',
    '--site-id',
    metavar='SITE_ID',
    type=int,
    help='Unique ID of the Site this Interface is under.  [required]',
    callback=callbacks.process_site_id,
)
@click.option(
    '-t',
    '--type',
    metavar='TYPE',
    type=int,
    help='The Interface type ID (e.g. 6 for ethernet).',
)
@click.pass_context
def add(ctx, attributes, addresses, device, description, mac_address,
        name, parent_id, speed, site_id, type):
    """
    Add a new Interface.

    You must provide a Device hostname or ID using the -D/--device option.

    When adding a new Interface, you must provide a value for the -n/--name
    option.

    If you wish to add attributes, you may specify the -a/--attributes
    option once for each key/value pair.

    If you wish to assign addresses, you may specify the -c/--addresses
    option once for each IP address.
    """
    data = ctx.params

    # Required option
    if name is None:
        raise click.UsageError('Missing option "-n" / "--name"')

    # Remove if empty; allow default assignment
    if speed is None:
        data.pop('speed')
    if type is None:
        data.pop('type')
    if description is None:
        data.pop('description')

    ctx.obj.add(data)


# List
@cli.group(invoke_without_command=True)
@click.option(
    '-a',
    '--attributes',
    metavar='ATTRS',
    help='A key/value pair attached to this Interface (format: key=value).',
    multiple=True,
)
@click.option(
    '-d',
    '--delimited',
    is_flag=True,
    help='Display set query results separated by commas vs. newlines.',
    default=False,
    show_default=True,
)
@click.option(
    '-D',
    '--device',
    metavar='DEVICE',
    type=types.NATURAL_KEY,
    help='Unique ID or hostname of the Device being retrieved.',
)
@click.option(
    '-e',
    '--description',
    metavar='DESCRIPTION',
    type=str,
    help='Filter by Interfaces matching this description.',
)
@click.option(
    '-g',
    '--grep',
    is_flag=True,
    help='Display list results in a grep-friendly format.',
    default=False,
    show_default=True,
)
@click.option(
    '-i',
    '--id',
    metavar='ID',
    help='Unique ID of the Interface being retrieved.',
)
@click.option(
    '-l',
    '--limit',
    metavar='LIMIT',
    help='Limit result to N resources.',
)
@click.option(
    '-m',
    '--mac-address',
    metavar='MAC',
    type=str,
    help='Filter by the MAC address of the Interface.',
)
@click.option(
    '-n',
    '--name',
    metavar='NAME',
    help='Filter to Interfaces matching this name.'
)
@click.option(
    '-N',
    '--natural-key',
    is_flag=True,
    help='Display list results by their natural key',
    default=False,
    show_default=True,
)
@click.option(
    '-o',
    '--offset',
    metavar='OFFSET',
    help='Skip the first N resources.',
)
@click.option(
    '-p',
    '--parent-id',
    metavar='PARENT_ID',
    type=types.NATURAL_KEY,
    help='Filter by integer of the ID of the parent Interface.',
)
@click.option(
    '-q',
    '--query',
    metavar='QUERY',
    help='Perform a set query using Attributes and output matching Interfaces.'
)
@click.option(
    '-S',
    '--speed',
    metavar='SPEED',
    type=int,
    help='Filter by integer of Mbps of interface (e.g. 20000 for 20 Gbps)',
)
@click.option(
    '-s',
    '--site-id',
    metavar='SITE_ID',
    help='Unique ID of the Site this Interface is under.  [required]',
    callback=callbacks.process_site_id,
)
@click.option(
    '-t',
    '--type',
    metavar='TYPE',
    type=int,
    help='Filter by integer of the interface type (e.g. 6 for ethernet)',
)
@click.pass_context
def list(ctx, attributes, delimited, device, description, grep, id, limit,
         mac_address, name, natural_key, offset, parent_id, query, site_id,
         speed, type):
    """
    List existing Interfaces for a Site.

    You must provide a Site ID using the -s/--site-id option.

    When listing Interfaces, all objects are displayed by default. You
    optionally may lookup a single Interfaces by ID using the -i/--id option.
    The ID can either be the numeric ID of the Interface, or the combination of
    the device's hostname and the Interface name separated by a colon.

    Example: switch-nyc3:Ethernet2

    You may limit the number of results using the -l/--limit option.

    You may look up the interfaces for a single Device using -D/--device which
    can be either a hostname or ID for a Device.
    """
    data = ctx.params
    data.pop('delimited')  # We don't want this going to the server.

    # If we provide ID, show more fields
    if id is not None or all([device, name]):
        display_fields = VERBOSE_FIELDS
    else:
        display_fields = DISPLAY_FIELDS

    # FIXME(jathan): If it's not a digit, it's a hostname? This is a hack for
    # the mixed use of natural_key vs id. We can do better "somehow".
    if device and not device.isdigit():
        log.debug('Device is hostname! Converting device => device_hostname')
        data['device_hostname'] = data.pop('device')

    # If we aren't passing a sub-command, just call list(), otherwise let it
    # fallback to default behavior.
    if ctx.invoked_subcommand is None:
        if query is not None:
            ctx.obj.natural_keys_by_query(data, delimited)
        else:
            ctx.obj.list(
                data, display_fields=display_fields,
                verbose_fields=VERBOSE_FIELDS
            )


@list.command()
@click.pass_context
def addresses(ctx, *args, **kwargs):
    """Get addresses assigned to an Interface."""
    callbacks.list_subcommand(
        ctx, display_fields=NETWORK_DISPLAY_FIELDS, my_name=ctx.info_name
    )


@list.command()
@click.pass_context
def parent(ctx, *args, **kwargs):
    """Get the parent Interface of an Interface."""
    callbacks.list_subcommand(
        ctx, display_fields=VERBOSE_FIELDS, my_name=ctx.info_name
    )


@list.command()
@click.pass_context
def ancestors(ctx, *args, **kwargs):
    """Recursively get all parents of an Interface."""
    callbacks.list_subcommand(
        ctx, display_fields=DISPLAY_FIELDS, my_name=ctx.info_name
    )


@list.command()
@click.pass_context
def children(ctx, *args, **kwargs):
    """Get immediate children of an Interface."""
    callbacks.list_subcommand(
        ctx, display_fields=DISPLAY_FIELDS, my_name=ctx.info_name
    )


@list.command()
@click.pass_context
def descendants(ctx, *args, **kwargs):
    """Recursively get all children of an Interface."""
    callbacks.list_subcommand(
        ctx, display_fields=DISPLAY_FIELDS, my_name=ctx.info_name
    )


@list.command()
@click.pass_context
def root(ctx, *args, **kwargs):
    """Get parent of all ancestors of an Interface."""
    callbacks.list_subcommand(
        ctx, display_fields=VERBOSE_FIELDS, my_name=ctx.info_name
    )


@list.command()
@click.pass_context
def siblings(ctx, *args, **kwargs):
    """Get Interfaces with the same parent and device id as an Interface."""
    callbacks.list_subcommand(
        ctx, display_fields=VERBOSE_FIELDS, my_name=ctx.info_name
    )


ASSIGNMENT_FIELDS = (
    ('id', 'ID'),
    ('hostname', 'Device'),
    ('device', 'Device ID'),
    ('address', 'Address'),
    ('interface_name', 'Interface'),
    ('interface', 'Interface ID'),
)


@list.command()
@click.pass_context
def assignments(ctx, *args, **kwargs):
    """Get assignment information for an Interface."""
    callbacks.list_subcommand(
        ctx, display_fields=ASSIGNMENT_FIELDS, my_name=ctx.info_name
    )


@list.command()
@click.pass_context
def networks(ctx, *args, **kwargs):
    """Get networks attached to Interface."""
    callbacks.list_subcommand(
        ctx, display_fields=NETWORK_DISPLAY_FIELDS, my_name=ctx.info_name
    )


# Remove
@cli.command()
@click.option(
    '-i',
    '--id',
    metavar='ID',
    help='Unique ID of the Interface being deleted.',
    required=True,
)
@click.option(
    '-s',
    '--site-id',
    metavar='SITE_ID',
    type=int,
    help='Unique ID of the Site this Interface is under.  [required]',
    callback=callbacks.process_site_id,
)
@click.pass_context
def remove(ctx, id, site_id):
    """
    Remove an Interface.

    You must provide a Site ID using the -s/--site-id option.

    When removing an Interface, you must provide the ID of the Interface using
    -i/--id. The ID can either be the numeric ID of the Interface or the
    combination of the device's hostname and Interfaces name separated by a
    colon.

    Example: switch-nyc3:Ethernet2

    You may retrieve the ID for an Interface by parsing it from the list of
    Interfaces for a given Site:

        nsot interfaces list --site-id <site_id> | grep <interface>
    """
    data = ctx.params
    ctx.obj.remove(**data)


# Update
@cli.command()
@click.option(
    '-a',
    '--attributes',
    metavar='ATTRS',
    help='A key/value pair attached to this Interface (format: key=value).',
    multiple=True,
    callback=callbacks.transform_attributes,
)
@click.option(
    '-c',
    '--addresses',
    metavar='ADDRS',
    help='An IPv4/IPv6 address in CIDR format.',
    multiple=True,
)
@click.option(
    '-e',
    '--description',
    metavar='DESCRIPTION',
    type=str,
    help='The description for this Interface.',
)
@click.option(
    '-i',
    '--id',
    metavar='ID',
    type=types.NATURAL_KEY,
    help='Unique ID of the Interface being updated.',
    required=True,
)
@click.option(
    '-m',
    '--mac-address',
    metavar='MAC',
    type=str,
    help='The MAC address of the Interface.',
)
@click.option(
    '-n',
    '--name',
    metavar='NAME',
    type=str,
    help='The name of the Interface.',
)
@click.option(
    '-p',
    '--parent-id',
    metavar='PARENT_ID',
    type=types.NATURAL_KEY,
    help='Unique ID of the parent interface.',
)
@click.option(
    '-S',
    '--speed',
    metavar='SPEED',
    type=int,
    help='The Interface speed in Mbps (e.g. 20000 for 20Gps).',
)
@click.option(
    '-s',
    '--site-id',
    metavar='SITE_ID',
    type=int,
    help='Unique ID of the Site this Interface is under.  [required]',
    callback=callbacks.process_site_id,
)
@click.option(
    '-t',
    '--type',
    metavar='TYPE',
    type=int,
    help='The Interface type ID (e.g. 6 for ethernet).',
)
@click.option(
    '--add-attributes',
    'attr_action',
    flag_value='add',
    default=True,
    help=(
        'Causes attributes to be added. This is the default and providing it '
        'will have no effect.'
    )
)
@click.option(
    '--delete-attributes',
    'attr_action',
    flag_value='delete',
    help=(
        'Causes attributes to be deleted instead of updated. If combined with'
        'with --multi the attribute will be deleted if either no value is '
        'provided, or if attribute no longer has an valid values.'
    ),
)
@click.option(
    '--replace-attributes',
    'attr_action',
    flag_value='replace',
    help=(
        'Causes attributes to be replaced instead of updated. If combined '
        'with --multi, the entire list will be replaced.'
    ),
)
@click.option(
    '--multi',
    is_flag=True,
    help='Treat the specified attributes as a list type.',
)
@click.pass_context
def update(ctx, attributes, addresses, description, id, mac_address, name,
           parent_id, speed, site_id, type, attr_action, multi):
    """
    Update an Interface.

    You must provide a Site ID using the -s/--site-id option.

    When updating an Interface you must provide the ID (-i/--id) and at least
    one of the optional arguments. The ID can either be the numeric ID of the
    Interface or the combination of the device's hostname and the interface
    name separated by a colon.

    Example: switch-nyc3:Ethernet2

    If you wish to assign addresses, you may specify the -c/--addresses option
    once for each IP address.

    The -a/--attributes option may be provided multiple times, once for each
    key-value pair.

    When modifying attributes you have three actions to choose from:

    * Add (--add-attributes). This is the default behavior that will add
    attributes if they don't exist, or update them if they do.

    * Delete (--delete-attributes). This will cause attributes to be
    deleted. If combined with --multi the attribute will be deleted if
    either no value is provided, or if the attribute no longer contains a
    valid value.

    * Replace (--replace-attributes). This will cause attributes to
    replaced. If combined with --multi and multiple attributes of the same
    name are provided, only the last value provided will be used.
    """
    if not any([name, addresses, attributes, description, mac_address,
                parent_id, speed, type]):
        msg = 'You must supply at least one of the optional arguments.'
        raise click.UsageError(msg)

    data = ctx.params
    ctx.obj.update(data)



# -*- coding: utf-8 -*-

'''API Model Classes

These resource models are derrived from collections.MutableMapping and, thus,
act like dicts and can instantiate with raw resource output from API as well as
simplifying by providing site_id and (usually) the natural key (cidr, hostname,
etc).

Example
-------

>>> from pynsot.models import Network, Device, Interface
>>> from pynsot.client import get_api_client
>>> client = get_api_client()
>>> net = Network(raw=client.networks.get()[-1])
>>>
>>> # Or also...
>>> # >>> net = Network(client=client, site_id=1, cidr='8.8.8.0/24')
>>>
>>> net.exists()
>>> True
>>>
>>> net.existing_resource()
{u'attributes': {},
 u'id': 81,
 u'ip_version': u'4',
 u'is_ip': False,
 u'network_address': u'254.0.0.0',
 u'parent_id': 1,
 u'prefix_length': 24,
 u'site_id': 1,
 u'state': u'allocated'}
>>>
>>> net.purge()
True
>>>
>>> net.exists()
False
>>>
>>> net.ensure()
True
>>>
>>> net.exists()
True
>>>
>>> net['site_id']
2
>>>
>>> net['site_id'] = 4
>>>
>>> net.exists()
False
>>>
>>> net['site_id'] = 2
>>>
>>> net.exists()
True
'''

from __future__ import unicode_literals
import logging
import collections
from abc import abstractproperty, abstractmethod, ABCMeta
from netaddr import IPNetwork
from pynsot.util import get_result
from pynsot.client import get_api_client


class Resource(collections.MutableMapping):
    '''Base API Abstraction Models Class

    Instances of an API abstraction model represent a single NSoT resource and
    provide methods for managing the state of it upstream. They can be
    instantiated by the raw returned object from NSoT API or more simply by a
    few descriptive kwargs.

    Resource is a subclass of :class:``collections.MutableMapping`` which makes
    it act as a dictionary would. The mapping represents the payload that would
    be accepted by NSoT and can be manipulated as desired like a normal dict.

    Subclassing:
        Subclasses must adhere to a simple contract:
            * Overload the abstractmethods and abstractproperties this class
              uses
            * If custom arguments are needed, overload ``self.postinit``.
              Just make sure to call ``self.init_payload()`` at the end. All
              kwargs that aren't handled by ``self.__init__`` are passed here

    :param site_id: Site ID this reseource belongs to. Required unless ``raw``
        is supplied.
    :type site_id: int
    :param attributes: Attributes to add to resource. If supplying ``raw``, add
        these after the instantiation like:

            >>> obj = Device(raw=RAW_API_DICT)
            >>> obj['attributes'] = {}

    :type attributes: dict
    :param client: Pynsot client for API interactions. Will be lazily loaded if
        not provided, but might be cheaper to supply it up front.
    :type client: pynsot.client.BaseClient
    :param raw: Raw NSoT resource object. What would be returned from a GET,
        POST, PUT, or PATCH operation for a single resource. Gets mapped
        directly to payload
    :type raw: dict
    '''

    __metaclass__ = ABCMeta

    def __init__(
        self,
        site_id=None,
        client=None,
        raw=None,
        attributes=None,
        **kwargs
    ):
        if raw is None:
            raw = {}
        if attributes is None:
            attributes = {}

        self.logger = logging.getLogger(__name__)
        self.errors = []
        self.last_error = None
        # Placeholder for .existing_resource() state
        self._existing_resource = {}
        self._payload = {}
        # Parameter validations

        # Site ID is required but can come from either kwarg or raw resource
        # dict, latter being preferred.
        if raw.get('site_id'):
            site_id = raw['site_id']
        elif site_id:
            pass  # Already set
        else:
            raise TypeError(
                'Resource requires site_id via param or ``raw`` key'
            )

        self._site_id = site_id
        self.client = client
        self.raw = raw
        self.attributes = attributes

        if not self.raw:
            self.postinit(**kwargs)
        else:
            # This is done here because subclases do this in their postinit().
            # If raw is passed, there's no reason to go through the postinit
            # process which takes things like network_address, hostname, etc.
            #
            # Every subclasses init_payload method has a condition for 'if raw
            # set raw' basically.
            self.init_payload()

    def postinit(self, **kwargs):
        '''Overloadable method for post __init__

        Use this for things that need to happen post-init, including
        subclass-specific argument handling.

        This method is called at the very end of __init__ unless ``raw`` is
        given.

        :params kwargs: All unhandled kwargs from __init__ are passed here
        :type kwargs: dict
        '''
        # If not being overloaded by subclass, still need to call this required
        # method if ``raw`` isn't provided
        self.init_payload()
        pass

    def ensure_client(self, **kwargs):
        '''Ensure that object has a client object

        Client may be passed during __init__ as a kwarg, but call this before
        doing any client work to ensure

        :param kwargs: These will be passed to get_api_client.
        :type kwargs: dict
        '''
        if self.client:
            return
        else:
            self.client = get_api_client(**kwargs)

    @abstractproperty
    def identifier(self):
        '''Human-friendly string to represent the resource

        Used in log messages and magic methods for comparison. Examples here
        would be CIDR, hostname, etc

        :rtype: str
        '''
        pass

    @property
    def resource(self):
        '''Pynsot client for resource type

        :rtype: pynsot.client.BaseClient
        '''
        self.ensure_client()
        return getattr(self.client, self.resource_name)

    @abstractproperty
    def resource_name(self):
        '''Name of resource

        Must be plural

        :rtype: str
        '''
        pass

    @abstractmethod
    def init_payload(self):
        '''
        Initializes the payload dictionary using resource specific data
        '''
        pass

    @property
    def payload(self):
        '''Represents exact payload sent to NSoT server

        :returns: _payload
        :rtype: dict
        '''
        return self._payload

    @payload.setter
    def payload(self, value):
        '''Setter for payload'''
        self._payload = value

    def __iter__(self):
        '''Iterate through payload keys'''
        return iter(self.payload)

    def __getitem__(self, key):
        '''Get item from payload'''
        return self.payload[key]

    def __setitem__(self, key, value):
        '''Set item in payload

        Cache is cleared here since the current resource has changed
        '''
        self.clear_cache()
        self.payload[key] = value

    def __delitem__(self, key):
        '''Delete item from payload

        Cache is cleared here since the current resource has changed
        '''
        self.clear_cache()
        del self.payload[key]

    def __len__(self):
        return len(self._payload)

    def __repr__(self):
        '''Human-friendly representation'''
        title = self.resource_name[:-1].title()
        return '<%s: %s>' % (title, str(self))

    def __str__(self):
        return str(self.identifier)

    def __eq__(self, other):
        '''Using ``identifier`` and ``site_id``, compare to another resource'''
        try:
            x = '%s:%s' % (self.identifier, self._site_id)
            y = '%s:%s' % (other.identifier, other._site_id)
            return x == y
        except:
            raise TypeError('Other object is not a Resource type')

    def log_error(self, error):
        '''Log and append errors to object

        This does a check to see if response object available from HTTP
        request. If not, that's OK too and it'll just append the raw error.
        '''
        if hasattr(error, 'response') and hasattr(error.response, 'json'):
            try:
                meta = error.response.json()
            except:
                meta = error.response
        else:
            meta = error

        self.logger.debug(
            '[%s] %s' % (self.identifier, meta),
            exc_info=True,
        )
        self.logger.warning(
            '[%s] %s' % (self.identifier, meta),
        )
        self.errors.append(meta)
        self.last_error = meta

    def existing_resource(self):
        '''Returns upstream resource

        If nothing exists, empty dict is returned. The result of this is cached
        in a property (_existing_resource) for cheaper re-checks. This can be
        cleared via ``.clear_cache()``

        :rtype: dict
        '''
        self.ensure_client()
        if self._existing_resource:
            return self._existing_resource
        else:
            cur = dict(self)
            cur.pop('attributes', None)
            # We pop attributes because the query fails leaving them in
            try:
                # Site client of resource type because NSoT doesn't support
                # passing site_id as a query parameter
                site = getattr(
                    self.client.sites(self['site_id']),
                    self.resource_name,
                )
                lookup = get_result(site.get(**cur))
            except Exception as e:
                self.log_error(e)
                # There might be a better way to do this. If the NSoT server is
                # unreachable or otherwise the lookup fails it'll log properly
                # and return as if nothing exists. If the resource actually
                # exists but due to reachability it couldn't confirm some
                # action may be unnecessarily took user-side.
                #
                # Honestly, though, I think this is a decent trade-off
                self._existing_resource = {}
                return self._existing_resource

            existing = len(lookup) > 0
            if existing:
                # This is where state will be kept for this
                self._existing_resource = lookup[0]
                return self._existing_resource
            else:
                self._existing_resource = {}
                return self._existing_resource

    def clear_cache(self):
        '''Clears state of certain properties

        This is ideally done during a write operation against the API, such as
        ``.purge()`` or ``.ensure()``. Helps prevent representing out-of-date
        information
        '''
        self._existing_resource = {}

    def exists(self):
        '''Does the current resource exist?

        :rtype: bool
        '''
        return bool(self.existing_resource())

    def ensure(self):
        '''Ensure object in current state exists in NSoT

        By site, make sure resource exists. True if it is or was able to get to
        the desired state, False if not and logged in ``last_error``.

        Having this operation be done individually for each resource
        has some pros and cons. Generally as long as the amount of items
        is small enough, it's not a huge difference but it is an extra
        HTTP request.

        Sending in bulk halts at the first error and fails the following
        so it requires more handling.

        Cache is cleared first thing and before return.

        :rtype: bool
        '''
        self.clear_cache()
        to_ensure = dict(self)

        try:
            # PATCH instead of POST
            if self.exists() and self.resource_name != 'interfaces':
                self.logger.debug('[%s] Patching', self.identifier)
                to_ensure['id'] = self.existing_resource()['id']
                self.resource.patch([to_ensure])
                self.logger.info('[%s] has been patched!', self.identifier)
                self.clear_cache()
                return True

            else:  # POST for initially creating
                self.logger.debug('[%s] Creating', self.identifier)
                self.resource.post([to_ensure])
                self.logger.info('[%s] has been created!', self.identifier)
                return True
        except Exception as e:
            self.log_error(e)
            self.clear_cache()
            return False

    def purge(self):
        '''Ensure resource doesn't exist upstream

        By site, make sure resource is deleted. True if it is or was able to
        get to the desired state, False if not and logged in ``last_error``.

        Cache is cleared first thing and before return.

        :rtype: bool
        '''
        self.clear_cache()
        try:
            if self.exists():
                self.logger.debug('[%s] Deleting', self.identifier)
                id_ = self.existing_resource()['id']
                self.resource(id_).delete()
                self.logger.info('[%s] has been deleted!', self.identifier)
                self.clear_cache()
                return True

            else:
                # We should return True to show the state is already how we
                # want it to be. False should be for unable to ensure resource
                # is gone
                return True
        except Exception as e:
            self.log_error(e)
            self.clear_cache()
            return False


class Network(Resource):
    '''Network API Abstraction Model

    Subclass of Resource.

    >>> n = Network(cidr='8.8.8.0/24', site_id=1)
    >>> n.exists()
    False
    >>> n.ensure()
    True
    >>> n.exists()
    True
    >>> n.existing_resource()
    {u'attributes': {},
     u'id': 31,
     u'ip_version': u'4',
     u'is_ip': True,
     u'network_address': u'8.8.8.0',
     u'parent_id': 1,
     u'prefix_length': 24,
     u'site_id': 1,
     u'state': u'assigned'}
    >>>
    >>> n.purge()
    True
    >>> n.exists()
    False
    >>> n.closest_parent()
    <Network: 8.8.0.0/16>

    :param cidr: CIDR for network
    '''

    def postinit(self, cidr=None):
        if not any([cidr, self.raw]):
            raise TypeError('Networks require a cidr')
        net = IPNetwork(cidr)

        ver = net.version
        self.network_address = str(net.network)
        self.prefix_length = int(net.prefixlen)
        self.is_host = ver == 4 and net.prefixlen == 32 or net.prefixlen == 128
        self.init_payload()

    @property
    def identifier(self):
        return '%s/%d' % (self['network_address'], self['prefix_length'])

    @property
    def resource_name(self):
        return 'networks'

    def init_payload(self):
        '''This will init the payload property'''
        if self.raw:
            self.payload = self.raw
            return

        self.payload = {
            'is_ip': self.is_host,
            'network_address': self.network_address,
            'prefix_length': self.prefix_length,
            'state': self.is_host and 'assigned' or 'allocated',
            'site_id': self._site_id,
            'attributes': self.attributes,
        }

    def closest_parent(self):
        '''Returns resource object of the closest parent network

        Empty dictionary if no parent network

        :returns: Parent resource
        :rtype: pynsot.models.Network or dict
        '''
        self.ensure_client()
        site = getattr(self.client.sites(self['site_id']), self.resource_name)
        cidr = '%s/%s' % (self['network_address'], self['prefix_length'])
        try:
            lookup = get_result(site(cidr).closest_parent.get())
            return Network(raw=lookup)
        except Exception as e:
            self.log_error(e)
            return {}

    def __len__(self):
        return self['prefix_length']


class Device(Resource):
    '''Device Resource

    Subclass of Resource.

    >>> dev = Device(hostname='router1-nyc', site_id=1)
    >>> dev.exists()
    False
    >>>
    >>> dev.ensure()
    True
    >>>
    >>> dev.existing_resource()
    {u'attributes': {}, u'hostname': u'router1-nyc', u'id': 1, u'site_id': 1}
    >>> dev.purge()
    True
    >>>
    >>> dev.exists()
    False

    :param hostname: Device hostname
    '''

    def postinit(self, hostname=None):
        if not any([hostname, self.raw]):
            raise TypeError('Devices require a hostname')
        self.hostname = hostname
        self.init_payload()

    @property
    def identifier(self):
        return self.hostname

    @property
    def resource_name(self):
        return 'devices'

    def init_payload(self):
        if self.raw:
            self.payload = self.raw
            return

        self.payload = {
            'hostname': self.hostname,
            'site_id': self._site_id,
            'attributes': self.attributes,
        }


class Interface(Resource):
    '''Interface Resource

    Subclass of Resource

    :param addresses: Addresses on interface
    :type addresses: list
    :param description: Interface description
    :param device: Required, device interface is on. TODO: broken as currently
        implemented but will soon reflect the following type
    :type device: pynsot.models.Device
    :param type: Interface type as described by SNMP IF-TYPE's
    :type type: int
    :param mac_address: MAC of interface
    :param name: Required, name of interface
    :param parent_id: ID of parent interface
    :type parent_id: int
    :param speed: Speed of interface
    :type speed: int
    '''

    def postinit(self, **kwargs):
        self.addresses = kwargs.get('addresses') or []
        self.description = kwargs.get('description') or ''
        self.device = kwargs.get('device') or 0
        self._original_device = kwargs.get('device') or None
        self.type = kwargs.get('type') or 6
        self.mac_address = kwargs.get('mac_address') or '00:00:00:00:00:00'
        self.name = kwargs.get('name') or None
        self.parent_id = kwargs.get('parent_id') or None
        self.speed = kwargs.get('speed') or 1000

        if not all([self.name, self.device]) or self.raw:
            raise TypeError('Interfaces require both a name and device!')

        self.attempt_device()
        self.init_payload()

    def attempt_device(self):
        '''Attempt to set ``device`` attribute to its ID if hostname was given

        If an ID was provided during init, this happens in init. If a hostname
        was provided, we need to take care of a couple things.

        1. Attempt to fetch ID for existing device by the hostname. If this
           works, use this ID

        2. Should #1 fail, set device id to 0 to signify this device doesn't
           exist yet, hoping this method will be called again when it's time to
           create. The thought here is if a user is instantiating lots of
           resources at the same time, there's a chance the device this relates
           to is going to be created before this one actually gets called upon.

           0 is used instead of a more descriptive message because should the
           device still not exist when executing this, at least a device
           doesn't exist error would be returned instead of expecting int not
           str.
        '''

        rerun = isinstance(self._original_device, str) and self.device == 0
        first_run = isinstance(self.device, str)
        # If equal to 0, means it had failed before
        if rerun or first_run:
            d = Device(
                client=self.client,
                site_id=self._site_id,
                hostname=self._original_device,
            )
            if d.exists():
                self.device = d.existing_resource()['id']
                return True
            else:
                self.device = 0
                return False

    @property
    def identifier(self):
        return '%s on %s' % (self.name, self.device)

    @property
    def resource_name(self):
        return 'interfaces'

    def init_payload(self):
        if self.raw:
            self.payload = self.raw
            return

        # TODO: This currently will only work on init and not later since
        # init_payload is only called once. This is left over from when it was
        # just payload
        #
        # attempt_device should instead set self.payload['device'] directly
        self.attempt_device()
        self.payload = {
            'addresses': self.addresses,
            'description': self.description,
            'device': self.device,
            'type': self.type,
            'mac_address': self.mac_address,
            'name': self.name,
            'parent_id': self.parent_id,
            'speed': self.speed,
            'attributes': self.attributes,
            'site_id': self._site_id,
        }



# -*- coding: utf-8 -*-

"""
Test the dotfile.
"""

from __future__ import unicode_literals
import copy
import logging
import os
import tempfile
import unittest

from pynsot import constants, dotfile

from .fixtures import DOTFILE_CONFIG_DATA


log = logging.getLogger(__name__)


class TestDotFile(unittest.TestCase):
    def setUp(self):
        """Automatically create a tempfile for each test."""
        fd, filepath = tempfile.mkstemp()
        self.filepath = filepath
        self.config_data = copy.deepcopy(DOTFILE_CONFIG_DATA)

        self.config_path = os.path.expanduser('~/.pynsotrc')
        self.backup_path = self.config_path + '.orig'
        self.backed_up = False
        if os.path.exists(self.config_path):
            log.debug('Config found, backing up...')
            os.rename(self.config_path, self.backup_path)
            self.backed_up = True

    def test_read_success(self):
        """Test that file can be read."""
        config = dotfile.Dotfile(self.filepath)
        config.write(self.config_data['auth_token'])
        config.read()

    def test_read_failure(self):
        """Test that that a missing file raises an error."""
        self.filepath = tempfile.mktemp()  # Doesn't create the temp file
        config = dotfile.Dotfile(self.filepath)
        self.assertRaises(
            IOError,  # This means it's trying to read from stdin
            config.read,
        )

    def test_write(self):
        """Test that file can be written."""
        config = dotfile.Dotfile(self.filepath)
        config.write(self.config_data['auth_token'])

    def test_validate_perms_success(self):
        """Test that file permissions are ok."""
        self.filepath = tempfile.mktemp()  # Doesn't create a temp file
        config = dotfile.Dotfile(self.filepath)
        config.write({})
        config.validate_perms()

    def test_validate_fields_auth_token(self):
        """Test that auth_token fields check out."""
        config = dotfile.Dotfile(self.filepath)

        self._validate_test_fields('auth_token', config)

    def test_validate_fields_auth_header(self):
        """Test that auth_header fields check out."""
        config = dotfile.Dotfile(self.filepath)

        self._validate_test_fields('auth_header', config)

    def _validate_test_fields(self, auth_method, config):
        config_data = self.config_data[auth_method]

        # We're not testing optional fields, yo.
        for optional_field in constants.OPTIONAL_FIELDS:
            config_data.pop(optional_field, None)

        # We're going to test every required field.
        fields = sorted(config_data)
        my_config = {}

        # Fetch the required_fields for this auth_method
        required_fields = config.get_required_fields(auth_method)

        # Iterate through the sorted list of fields to make sure that each one
        # raises an error as expected.
        err = 'Missing required field: '
        for field in fields:
            with self.assertRaisesRegexp(dotfile.DotfileError, err + field):
                config.read()
                config.validate_fields(my_config, required_fields)

            my_config[field] = config_data[field]
            config.write(my_config)

    def tearDown(self):
        """Delete a tempfile if it exists. Otherwise carry on."""
        try:
            os.remove(self.filepath)
        except OSError:
            pass

        if self.backed_up:
            log.debug('Restoring original config.')
            os.rename(self.backup_path, self.config_path)  # Restore original



# -*- coding: utf-8 -*-

"""
Constant values used across the project.
"""

from __future__ import unicode_literals
import os


__author__ = 'Jathan McCollum'
__maintainer__ = 'Jathan McCollum'
__email__ = 'jathan@dropbox.com'
__copyright__ = 'Copyright (c) 2016 Dropbox, Inc.'


# Header used for passthrough authentication.
AUTH_HEADER = 'X-NSoT-Email'

# Default authentication method
DEFAULT_AUTH_METHOD = 'auth_token'

# Mapping of required field names and default values we want to be in the
# dotfile.
REQUIRED_FIELDS = {
    'auth_method': ['auth_token', 'auth_header'],
    'url': None,
}

# Fields that map to specific auth_methods.
SPECIFIC_FIELDS = {
    'auth_header': {
        'default_domain': 'localhost',
        'auth_header': AUTH_HEADER,
    }
}

# Mapping of optional field names and default values (if any)
OPTIONAL_FIELDS = {
    'default_site': None,
    'api_version': None,
}

# Path stuff
USER_HOME = os.path.expanduser('~')
DOTFILE_NAME = '.pynsotrc'
DOTFILE_PATH = os.path.join(USER_HOME, DOTFILE_NAME)
DOTFILE_PERMS = 0600  # -rw-------

# Config section name
SECTION_NAME = 'pynsot'



# -*- coding: utf-8 -*-

from __future__ import absolute_import, unicode_literals

import pytest

from tests.fixtures import (protocol_type, site_client)

from .circuits import (circuit, device_a, interface_a)

@pytest.fixture
def protocol(site_client, device_a, interface_a, circuit, protocol_type):
    """
    Return a Protocol Object.
    """
    device_id = device_a['id']
    interface_slug = interface_a['name_slug']
    return site_client.sites(site_client.default_site).protocols.post(
        {
            'device': device_id,
            'type': 'bgp',
            'interface': interface_slug,
            'attributes': {'foo': 'test_protocol'},
            'circuit': circuit['name'],
            'description': 'bgp is the best',
        }
    )

@pytest.fixture
def protocol_attribute(site_client):
    return site_client.sites(site_client.default_site).attributes.post(
        {
            'name':'boo',
            'value': 'test_attribute',
            'resource_name': 'Protocol',
        }
    )

@pytest.fixture
def protocol_attribute2(site_client):
    return site_client.sites(site_client.default_site).attributes.post(
        {
            'name':'foo',
            'value': 'test_protocol',
            'resource_name': 'Protocol',
        }
    )


@pytest.fixture
def protocols(site_client, protocol_type, protocol_attribute2):
    """
    A group of Protocol objects for testing limit/offest/etc.
    """
    protocols = []
    site = site_client.sites(site_client.default_site)
    for i in range(1, 6):
        device_name = 'device{:02d}'.format(i)
        interface_name = 'Ethernet1/{}'.format(i)

        device = site.devices.post({'hostname': device_name})
        interface = site.interfaces.post({
            'device': device['id'],
            'name': interface_name,
        })

        protocol = site.protocols.post({
            'type': protocol_type['name'],
            'device': device['id'],
            'interface': interface['id'],
            'attributes': {'foo': 'bar'},
        })
        protocols.append(protocol)

    return protocols



# -*- coding: utf-8 -*-

"""
Test the CLI app.
"""

from __future__ import unicode_literals
import logging

import pytest

from .fixtures import (attribute, attributes, client, config, device, network,
                       interface, site, site_client)
from .util import CliRunner, assert_output


log = logging.getLogger(__name__)


__all__ = ('client', 'config', 'site', 'site_client', 'pytest', 'attribute',
           'device', 'interface', 'network')


#########
# Sites #
#########
def test_site_id(client):
    """Test ``nsot devices list`` without required site_id"""
    runner = CliRunner(client.config)
    with runner.isolated_filesystem():
        result = runner.run('devices list')

        # Make sure it says site-id is required
        expected_output = 'Error: Missing option "-s" / "--site-id".'
        assert result.exit_code == 2
        assert expected_output in result.output


def test_site_add(client):
    """Test ``nsot sites add``."""
    runner = CliRunner(client.config)
    with runner.isolated_filesystem():
        # Make sure it is a positive confirmation.
        result = runner.run("sites add -n Foo -d 'Foo site.'")
        expected_output = "[SUCCESS] Added site!\n"
        assert result.exit_code == 0
        assert result.output == expected_output

        # Try to add the same site again and fail.
        result = runner.run("sites add -n Foo -d 'Foo site.'")
        expected_output = 'site with this name already exists.\n'
        assert result.exit_code == 1
        assert expected_output in result.output


def test_sites_list(client, site):
    """Test ``nsot sites list``."""
    runner = CliRunner(client.config)
    with runner.isolated_filesystem():
        # Simply list the site successfully.
        result = runner.run('sites list')
        assert result.exit_code == 0
        assert site['name'] in result.output

        # Test -i/--id
        result = runner.run('sites list -i %s' % site['id'])
        assert result.exit_code == 0
        assert site['name'] in result.output

        # Test -n/--name
        result = runner.run('sites list -n %s' % site['name'])
        assert result.exit_code == 0
        assert site['name'] in result.output

        # Test -N/--natural-key
        result = runner.run('sites list -N')
        assert result.exit_code == 0
        assert site['name'] == result.output.strip()


def test_sites_update(client, site):
    """Test ``nsot sites update``."""
    runner = CliRunner(client.config)
    with runner.isolated_filesystem():
        # Change the name.
        result = runner.run('sites update -n Bacon -i %s' % site['id'])
        assert result.exit_code == 0
        assert 'Updated site!' in result.output

        # Update the description
        result = runner.run('sites update -d Sizzle -i %s' % site['id'])
        assert result.exit_code == 0
        assert 'Updated site!' in result.output

        # Assert the bacon sizzles
        result = runner.run('sites list -n Bacon')
        assert result.exit_code == 0
        assert 'Bacon' in result.output
        assert 'Sizzle' in result.output


def test_sites_remove(client, site):
    """Test ``nsot sites remove``."""
    runner = CliRunner(client.config)
    with runner.isolated_filesystem():
        # Just delete the site we have.
        result = runner.run('sites remove -i %s' % site['id'])
        assert result.exit_code == 0
        assert 'Removed site!' in result.output


##############
# Attributes #
##############
def test_attributes_add(site_client):
    """Test ``nsot attributes add``."""
    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        # Create a new attribute
        result = runner.run(
            'attributes add -n device_multi -r device --multi'
        )
        assert_output(result, ['Added attribute!'])


def test_attributes_list(site_client):
    """Test ``nsot attributes list``."""
    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        # Create the monitored attribute
        runner.run('attributes add -n monitored -r device --allow-empty')

        # Simple list
        result = runner.run('attributes list')
        assert result.exit_code == 0

        # Test -N/--natural-key
        result = runner.run('attributes list -N')
        assert result.exit_code == 0
        assert 'Device:monitored\n' == result.output

        # List a single attribute by name
        attr = site_client.attributes.get(name='monitored')[0]
        name_result = runner.run('attributes list -n monitored')

        # Single matching object should have 'Constraints' column
        expected = ('Constraints', 'monitored')
        assert result.exit_code == 0
        for e in expected:
            assert e in name_result.output

        # List the same attribute by id
        id_result = runner.run('attributes list -i %s' % attr['id'])
        assert id_result.exit_code == 0

        # Output should match the previous command.
        assert id_result.output == name_result.output


def test_attributes_update(site_client):
    """Test ``nsot attributes update``."""
    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        # Create and retrieve the 'tags' attribute as a list type
        runner.run('attributes add -r device -n tags --multi')
        attr = site_client.attributes.get(name='tags')[0]

        # Display the attribute before update.
        before_result = runner.run('attributes list -i %s' % attr['id'])
        assert_output(before_result, ['tags', 'Device'])

        # Update the tags attribute to disable multi
        result = runner.run('attributes update --no-multi -i %s' % attr['id'])
        assert_output(result, ['Updated attribute!'])

        # List it to show the proof that the results are not the same.
        after_result = runner.run('attributes list -i %s' % attr['id'])
        assert after_result.exit_code == 0
        assert before_result != after_result

        # Update attribute by natural_key (name, resource_name)
        runner.run(
            'attributes update -r device -n tags --allow-empty'
        )
        result = runner.run('attributes list -r device -n tags')
        assert_output(result, ['allow_empty=True'])

        # Run update without optional args
        result = runner.run('attributes update -r device -n tags')
        assert_output(result, ['Error:'], exit_code=2)


def test_attributes_remove(site_client, attribute):
    """Test ``nsot attributes update``."""
    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        # Just delete the attribute we have.
        result = runner.run('attributes remove -i %s' % attribute['id'])
        assert result.exit_code == 0
        assert 'Removed attribute!' in result.output


###########
# Devices #
###########
def test_device_add(site_client):
    """Test ``nsot devices add``."""
    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        # Success is fun!
        result = runner.run('devices add -H foo-bar1')
        expected_output = '[SUCCESS] Added device!\n'
        assert result.exit_code == 0
        assert result.output == expected_output


def test_devices_bulk_add(site_client):
    """Test ``nsot devices add -b /path/to/bulk_file``"""
    BULK_ADD = (
        'hostname:attributes\n'
        'foo-bar1:owner=jathan\n'
        'foo-bar2:owner=jathan\n'
    )

    # This has an invalid attribute (bacon)
    BULK_FAIL = (
        'hostname:attributes\n'
        'foo-bar3:owner=jathan,bacon=delicious\n'
        'foo-bar4:owner=jathan\n'
    )

    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        # Create the attribute
        runner.run('attributes add -n owner -r device')

        # Write the bulk files.
        with open('bulk_file', 'w') as fh:
            fh.writelines(BULK_ADD)
        with open('bulk_fail', 'w') as fh:
            fh.writelines(BULK_FAIL)

        # Test valid bulk_add
        result = runner.run('devices add -b bulk_file')
        expected_output = (
            "[SUCCESS] Added device!\n"
            "[SUCCESS] Added device!\n"
        )
        assert result.exit_code == 0
        assert result.output == expected_output

        # Test an invalid add
        result = runner.run('devices add -b bulk_fail')
        expected_output = 'Attribute name (bacon) does not exist'
        assert result.exit_code == 1
        assert expected_output in result.output


def test_devices_list(site_client):
    """Test ``nsot devices list``."""
    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        # Create the owner attribute
        runner.run('attributes add -n owner -r device')

        # Create 2 devices w/ owner= set
        runner.run('devices add -H foo-bar1 -a owner=jathan')
        runner.run('devices add -H foo-bar2 -a owner=jathan')

        # Make sure the hostnames show up in a normal list
        result = runner.run('devices list')
        assert result.exit_code == 0

        expected = ('foo-bar1', 'foo-bar2')
        for e in expected:
            assert e in result.output

        # Set query display newline-delimited (default)
        result = runner.run('devices list -q owner=jathan')
        expected_output = (
            'foo-bar1\n'
            'foo-bar2\n'
        )
        assert result.exit_code == 0
        assert result.output == expected_output

        # Test -N/--natural-key
        result = runner.run('devices list -N')
        assert result.exit_code == 0
        assert result.output == expected_output  # Same output as above

        # Set query display comma-delimited (-d/--delimited)
        result = runner.run('devices list -q owner=jathan -d')
        expected_output = 'foo-bar1,foo-bar2\n'
        assert result.exit_code == 0
        assert result.output == expected_output

        # Set query with --l/--limit
        result = runner.run('devices list -l 1 -q owner=jathan')
        expected_output = 'foo-bar1\n'
        assert result.exit_code == 0
        assert result.output == expected_output

        # Set query with --l/--limit and -o/--offset
        result = runner.run('devices list -l 1 -o 1 -q owner=jathan')
        expected_output = 'foo-bar2\n'
        assert result.exit_code == 0
        assert result.output == expected_output

        # Grep-friendly output (-g/--grep)
        result = runner.run('devices list -a owner=jathan -g')
        expected_output = (
            u'foo-bar1 owner=jathan\n'
            'foo-bar1 hostname=foo-bar1\n'
            'foo-bar1 id=5\n'
            'foo-bar1 site_id=11\n'
            'foo-bar2 owner=jathan\n'
            'foo-bar2 hostname=foo-bar2\n'
            'foo-bar2 id=6\n'
            'foo-bar2 site_id=11\n')

        assert result.exit_code == 0
        assert result.output == expected_output

        # Now create 1 device w/ owner= w/ a space in the value
        runner.run('devices add -H foo-bar3 -a owner="Jathan McCollum"')

        # Test that you can query by values w/ spaces when properly quoted
        result = runner.run('devices list -q \'owner="Jathan McCollum"\'')
        expected_output = 'foo-bar3\n'
        assert result.exit_code == 0
        assert result.output == expected_output

        # ... Or using backslashes works, too.
        result = runner.run('devices list -q "owner=Jathan\ McCollum"')
        assert result.exit_code == 0
        assert result.output == expected_output

        # Test that query with unbalanced quotes fails.
        result = runner.run('devices list -q \'owner="Jathan McCollum\'')
        assert result.exit_code == 1
        assert 'No closing quotation' in result.output


def test_devices_subcommands(site_client, device):
    """Test ``nsot devices list ... interfaces`` sub-command."""
    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        # Create two interfaces on the device.
        hostname = device['hostname']
        device_id = device['id']
        runner.run('interfaces add -D %s -n eth0' % device_id)
        runner.run('interfaces add -D %s -n eth1' % device_id)

        # Lookup using natural_key (hostname)
        result = runner.run('devices list -H %s interfaces' % hostname)
        expected = ('eth0', 'eth1')
        assert result.exit_code == 0
        for e in expected:
            assert e in result.output

        # Lookup by id
        result = runner.run('devices list -i %s interfaces' % device_id)
        assert result.exit_code == 0
        for e in expected:
            assert e in result.output

        # Lookup by set query
        result = runner.run('devices list -q foo=test_device interfaces')
        assert result.exit_code == 0
        for e in expected:
            assert e in result.output


def test_devices_update(site_client):
    """Test ``nsot devices update``."""
    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        # Create the attributes
        runner.run('attributes add -n owner -r device')
        runner.run('attributes add -n monitored -r device --allow-empty')

        # Create the device w/ owner= set
        runner.run('devices add -H foo-bar1 -a owner=jathan')

        # Now set the 'monitored' attribute
        result = runner.run('devices update -H foo-bar1 -a monitored')
        expected_output = "[SUCCESS] Updated device!\n"
        assert result.exit_code == 0
        assert result.output == expected_output

        # Run a list to assert 'monitored=' attribute is now there.
        result = runner.run('devices list -H foo-bar1')
        assert result.exit_code == 0
        assert 'monitored=' in result.output

        # Now run update by natural_key (hostname) to remove monitored
        result = runner.run(
            'devices update -H foo-bar1 -a monitored --delete-attributes'
        )
        assert result.exit_code == 0

        # Make sure that monitored isn't showing in output
        result = runner.run('devices list -H foo-bar1')
        assert result.exit_code == 0
        assert 'monitored=' not in result.output


def test_attribute_modify_multi(site_client):
    """Test modification of list-type attributes (multi=True)."""
    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        #####
        # ADD a multi attribute with 2 items
        #####
        # Create the initial device and multi attribute
        runner.run('devices add -H foo-bar1')
        runner.run('attributes add -r device -n multi --multi')

        # Add multi=[jathy, jilli]
        result = runner.run(
            'devices update -H foo-bar1 -a multi=jathy -a multi=jilli --multi'
        )
        assert result.exit_code == 0

        # List to see that multi= shows w/ jathy & jilli as values.
        result = runner.run('devices list -H foo-bar1')
        expected = ('multi=', 'jathy', 'jilli')
        assert result.exit_code == 0
        for e in expected:
            assert e in result.output

        #########
        # REPLACE it with two different items
        #########
        # Replace with multi=[bob, alice]
        result = runner.run(
            'devices update -H foo-bar1 -a multi=bob -a multi=alice --multi '
            '--replace-attributes'
        )
        assert result.exit_code == 0

        # List to show that multi= shows w/ bob & alice as values.
        result = runner.run('devices list -H foo-bar1')
        expected = ('multi=', 'bob', 'alice')
        assert result.exit_code == 0
        for e in expected:
            assert e in result.output

        ########
        # DELETE one, leaving one
        ########
        # Pop bob
        result = runner.run(
            'devices update -H foo-bar1 -a multi=bob --multi '
            '--delete-attributes'
        )
        assert result.exit_code == 0

        # List to show the proof that bob no longer shows!
        result = runner.run('devices list -H foo-bar1')
        assert result.exit_code == 0
        assert 'bob' not in result.output

        ########
        # DELETE the other; attr goes away, object returned to initial state
        ########

        # Pop alice; attribute is removed
        result = runner.run(
            'devices update -H foo-bar1 -a multi=alice --multi '
            '--delete-attributes'
        )
        assert result.exit_code == 0

        # List to show the proof that 'multi=' no longer shows in output!
        result = runner.run('devices list -H foo-bar1')
        assert result.exit_code == 0
        assert 'multi=' not in result.output

        #####
        # ADD new list w/ 2 items
        #####
        # Add multi=[spam, eggs]
        result = runner.run(
            'devices update -H foo-bar1 -a multi=spam -a multi=eggs --multi'
        )
        assert result.exit_code == 0

        # List to show that 'multi=' is back and has egg & spam.
        result = runner.run('devices list -H foo-bar1')
        expected = ('multi=', 'eggs', 'spam')
        assert result.exit_code == 0
        for e in expected:
            assert e in result.output

        ########
        # DELETE with no value; attribute goes away; object initialized
        ########
        result = runner.run(
            'devices update -H foo-bar1 -a multi --delete-attributes'
        )
        assert result.exit_code == 0

        # List to show the proof that 'multi=' no longer shows in output. And
        # scene.
        result = runner.run('devices list -H foo-bar1')
        assert result.exit_code == 0
        assert 'multi=' not in result.output


def test_devices_remove(site_client, device):
    """Test ``nsot devices remove``."""
    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        # Just delete the device we have.
        result = runner.run('devices remove -i %s' % device['id'])
        assert_output(result, ['Removed device!'])

        # Create another device and delete it by hostname using -H
        runner.run('devices add -H delete-me')
        result = runner.run('devices remove -i delete-me')
        assert_output(result, ['Removed device!'])

        # Create another device and delete it by hostname using -H
        runner.run('devices add -H delete-me')
        result = runner.run('devices remove -H delete-me')
        assert_output(result, ['Removed device!'])


############
# Networks #
############
def test_networks_add(site_client):
    """Test ``nsot networks add``."""
    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        result = runner.run('networks add -c 10.0.0.0/8')
        expected_output = '[SUCCESS] Added network!\n'
        assert result.exit_code == 0
        assert result.output == expected_output


def test_networks_bulk_add(site_client):
    """Test ``nsot networks add -b /path/to/bulk_file``."""
    BULK_ADD = (
        'cidr:attributes\n'
        '10.0.0.0/8:owner=jathan\n'
        '10.0.0.0/24:owner=jathan\n'
    )
    BULK_FAIL = (
        'cidr:attributes\n'
        '10.10.0.0/24:owner=jathan,bacon=delicious\n'
        '10.11.0.0/24:owner=jathan\n'
    )

    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        # Create the owner attribute
        runner.run('attributes add -n owner -r network')

        # Write the bulk files.
        with open('bulk_file', 'w') as fh:
            fh.writelines(BULK_ADD)
        with open('bulk_fail', 'w') as fh:
            fh.writelines(BULK_FAIL)

        # Test *with* provided site_id
        result = runner.run('networks add -b bulk_file')
        expected_output = (
            "[SUCCESS] Added network!\n"
            "[SUCCESS] Added network!\n"
        )
        assert result.exit_code == 0
        assert result.output == expected_output

        # Test an invalid add
        result = runner.run('networks add -b bulk_fail')
        expected_output = 'Attribute name (bacon) does not exist'
        assert result.exit_code == 1
        assert expected_output in result.output


def test_networks_list(site_client):
    """Test ``nsot networks list``."""
    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        # Create the owner attribute
        runner.run('attributes add -n owner -r network')

        # First create our networks w/ owner= set
        runner.run('networks add -c 10.0.0.0/8 -a owner=jathan')
        runner.run('networks add -c 10.0.0.0/24 -a owner=jathan')

        # Make sure 10.0.0.0 shows twice in the output. Lazy man's output
        # checking.
        result = runner.run('networks list')
        assert result.output.count('10.0.0.0') == 3
        assert result.exit_code == 0

        # Set query display newline-delimited (default)
        result = runner.run('networks list -q owner=jathan')
        expected_output = (
            '10.0.0.0/8\n'
            '10.0.0.0/24\n'
        )
        assert result.exit_code == 0
        assert result.output == expected_output

        # Test -N/--natural-key
        result = runner.run('networks list -N')
        assert result.exit_code == 0
        assert result.output == expected_output  # Same output as above

        # Set query w/ -l/--limit
        result = runner.run('networks list -l 1 -q owner=jathan')
        expected_output = '10.0.0.0/8\n'
        assert result.exit_code == 0
        assert result.output == expected_output

        # Set query w/ -l/--limit & -o/--offset
        result = runner.run('networks list -l 1 -o 1 -q owner=jathan')
        expected_output = '10.0.0.0/24\n'
        assert result.exit_code == 0
        assert result.output == expected_output

        # Set query display comma-delimited (-d/--delimited)
        result = runner.run('networks list -q owner=jathan -d')
        expected_output = '10.0.0.0/8,10.0.0.0/24\n'
        assert result.exit_code == 0
        assert result.output == expected_output

        # Set query display grep-friendly (--g/--grep)
        result = runner.run('networks list -a owner=jathan -g')
        expected_output = (
            '10.0.0.0/8 owner=jathan\n'
            '10.0.0.0/8 cidr=10.0.0.0/8\n'
            '10.0.0.0/8 id=5\n'
            '10.0.0.0/8 ip_version=4\n'
            '10.0.0.0/8 is_ip=False\n'
            '10.0.0.0/8 network_address=10.0.0.0\n'
            '10.0.0.0/8 parent=None\n'
            '10.0.0.0/8 parent_id=None\n'
            '10.0.0.0/8 prefix_length=8\n'
            '10.0.0.0/8 site_id=18\n'
            '10.0.0.0/8 state=allocated\n'
            '10.0.0.0/24 owner=jathan\n'
            '10.0.0.0/24 cidr=10.0.0.0/24\n'
            '10.0.0.0/24 id=6\n'
            '10.0.0.0/24 ip_version=4\n'
            '10.0.0.0/24 is_ip=False\n'
            '10.0.0.0/24 network_address=10.0.0.0\n'
            '10.0.0.0/24 parent=10.0.0.0/8\n'
            '10.0.0.0/24 parent_id=5\n'
            '10.0.0.0/24 prefix_length=24\n'
            '10.0.0.0/24 site_id=18\n'
            '10.0.0.0/24 state=allocated\n'
        )
        assert result.exit_code == 0
        assert result.output == expected_output

        # Now create 1 network w/ owner= w/ a space in the value
        runner.run('networks add -c 10.0.0.0/16 -a owner="Jathan McCollum"')

        # Test that you can query by values w/ spaces when properly quoted
        result = runner.run('networks list -q \'owner="Jathan McCollum"\'')
        expected_output = '10.0.0.0/16\n'
        assert result.exit_code == 0
        assert result.output == expected_output

        # ... Or using backslashes works, too.
        result = runner.run('networks list -q "owner=Jathan\ McCollum"')
        assert result.exit_code == 0
        assert result.output == expected_output

        # Test that query with unbalanced quotes fails.
        result = runner.run('networks list -q \'owner="Jathan McCollum\'')
        assert result.exit_code == 1
        assert 'No closing quotation' in result.output


def test_networks_subcommands(site_client, network):
    """Test ``nsot networks list ... <subcommand>``."""
    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        # Create the owner attribute
        runner.run('attributes add -n owner -r network')

        # Create our networks w/ owner= set
        runner.run('networks add -c 10.0.0.0/8 -a owner=jathan')
        runner.run('networks add -c 10.0.0.0/24 -a owner=jathan')

        # Test subnets: Assert that 10.0.0.0/24 shows in output
        result = runner.run('networks list -c 10.0.0.0/8 subnets')
        assert_output(result, ['10.0.0.0', '24'])

        # Test supernets: Assert that 10.0.0.0/8 shows in output
        result = runner.run('networks list -c 10.0.0.0/24 supernets')
        assert_output(result, ['10.0.0.0', '8'])

        # Let's add some more networks for fun.
        runner.run('networks add -c 10.10.10.0/24')
        runner.run('networks add -c 10.10.10.1/32')
        runner.run('networks add -c 10.10.10.2/32')
        runner.run('networks add -c 10.10.10.3/32')

        # Test parent
        result = runner.run('networks list -c 10.10.10.1/32 parent')
        assert_output(result, ['10.10.10.0', '24'])

        # Test ancestors
        result = runner.run('networks list -c 10.10.10.1/32 ancestors')
        assert_output(result, ['10.10.10.0', '24'])
        assert_output(result, ['10.0.0.0', '8'])

        # Test children
        result = runner.run('networks list -c 10.10.10.0/24 children')
        assert_output(result, ['10.10.10.1', '32'])
        assert_output(result, ['10.10.10.2', '32'])
        assert_output(result, ['10.10.10.3', '32'])

        # Test descendants
        result = runner.run('networks list -c 10.0.0.0/8 descendants')
        assert_output(result, ['10.0.0.0', '24'])
        assert_output(result, ['10.10.10.0', '24'])
        assert_output(result, ['10.10.10.1', '32'])
        assert_output(result, ['10.10.10.2', '32'])
        assert_output(result, ['10.10.10.3', '32'])

        # Assert descendents (typoed) includes deprecation warning
        result2 = runner.run('networks list -c 10.0.0.0/8 descendents')
        assert_output(result2, ['[WARNING]'])
        assert result.output.splitlines() == result2.output.splitlines()[1:]

        # Test root
        result = runner.run('networks list -c 10.10.10.1/32 root')
        assert_output(result, ['10.0.0.0', '8'])

        # Test siblings
        result = runner.run('networks list -c 10.10.10.2/32 siblings')
        assert_output(result, ['10.10.10.1', '32'])
        assert_output(result, ['10.10.10.3', '32'])

        # Test siblings w/ --include-self
        result = runner.run(
            'networks list -c 10.10.10.2/32 siblings --include-self'
        )
        assert_output(result, ['10.10.10.1', '32'])
        assert_output(result, ['10.10.10.2', '32'])
        assert_output(result, ['10.10.10.3', '32'])

        # Test closest_parent PASS w/ non-existent network
        result = runner.run('networks list -c 10.10.10.104/32 closest_parent')
        assert_output(result, ['10.10.10.0', '24'])

        # Test closest_parent FAIL w/ non-existent parent
        result = runner.run('networks list -c 1.2.3.4/32 closest_parent')
        assert_output(result, ['No such Network found'], exit_code=1)

        # Test that subcommands work when given a query
        result = runner.run('networks list -q foo=test_network parent')
        assert result.exit_code == 0

        # Queries that return >1 Network should cause an error
        result = runner.run('networks list -q owner=jathan parent')
        assert result.exit_code == 1


def test_networks_allocation(site_client, device, network, interface):
    """Test network allocation-related subcommands."""
    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        # network = 10.20.30.0/24
        # leaf = 10.20.30.1/32

        # Test assignments
        result = runner.run('networks list -c 10.20.30.1/32 assignments')
        assert_output(result, ['foo-bar1', 'eth0'])

        # Test reserved
        runner.run('networks add -c 10.20.30.104/32 --state reserved')
        result = runner.run('networks list reserved')
        assert_output(result, ['10.20.30.104', '32'])

        # Test next_network
        result = runner.run(
            'networks list -c 10.20.30.0/24 next_network -n 2 -p 28'
        )
        assert_output(result, ['10.20.30.16', '28'])
        assert_output(result, ['10.20.30.32', '28'])

        # Test next_address
        runner.run('networks add -c 10.20.30.3/32')
        result = runner.run('networks list -c 10.20.30.0/24 next_address -n 3')
        assert_output(result, ['10.20.30.2', '32'])
        assert_output(result, ['10.20.30.4', '32'])
        assert_output(result, ['10.20.30.5', '32'])

        # Test strict allocations
        runner.run('networks add -c 10.2.1.0/24')
        runner.run('networks add -c 10.2.1.0/25')
        result = runner.run(
            'networks list -c 10.2.1.0/24 next_network -p 28 -n 3 -s')
        assert_output(result, ['10.2.1.128', '28'])
        assert_output(result, ['10.2.1.144', '28'])
        assert_output(result, ['10.2.1.160', '28'])

        # Test strict allocations for next_address
        result = runner.run(
            'networks list -c 10.2.1.0/24 next_address -n 3 -s')
        assert_output(result, ['10.2.1.128', '32'])
        assert_output(result, ['10.2.1.129', '32'])
        assert_output(result, ['10.2.1.130', '32'])


def test_networks_update(site_client):
    """Test ``nsot networks update``."""
    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        # Create the owner attribute
        runner.run('attributes add -n owner -r network')

        # Create a network & attribute
        runner.run('networks add -c 10.0.0.0/8 -a owner=jathan')
        runner.run('attributes add -n foo -r network')

        # Run the update to add the new attribute
        result = runner.run('networks update -c 10.0.0.0/8 -a foo=bar')
        expected_output = "[SUCCESS] Updated network!\n"
        assert result.exit_code == 0
        assert result.output == expected_output

        # Run a list to see the object w/ the updated result
        result = runner.run('networks list -c 10.0.0.0/8')
        assert result.exit_code == 0
        assert 'foo=bar' in result.output

        # Now run update by natural_key (cidr) to remove foo=bar
        result = runner.run(
            'networks update -c 10.0.0.0/8 -a foo --delete-attributes'
        )
        assert result.exit_code == 0
        assert 'foo=bar' not in result.output


def test_networks_remove(site_client, network):
    """Test ``nsot networks remove``."""
    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        # Just delete the network we have by id.
        result = runner.run('networks remove -i %s' % network['id'])
        assert_output(result, ['Removed network!'])

        # Create a new network and then delete it by CIDR using -i.
        runner.run('networks add -c 10.20.30.0/24')
        result = runner.run('networks remove -i 10.20.30.0/24')
        assert_output(result, ['Removed network!'])

        # Create a another network and then delete it by CIDR using -c.
        runner.run('networks add -c 10.20.30.0/24')
        result = runner.run('networks remove -c 10.20.30.0/24')
        assert_output(result, ['Removed network!'])

        # Create three networks, each children of the next.
        runner.run('networks add -c 10.20.30.0/24')
        runner.run('networks add -c 10.20.30.5/32')
        runner.run('networks add -c 10.20.0.0/17')

        # Delete middle parent without force-delete flag.
        result = runner.run('networks remove -c 10.20.30.0/24')
        result.exit_code == 1
        assert "Cannot delete some instances of model 'Network'" in result.output

        # Delete middle parent with force-delete flag.
        result = runner.run('networks remove -c 10.20.30.0/24 -f')
        assert_output(result, ['Removed network!'])

        # Delete parent with child nodes using force-delete flag, it will fail.
        result = runner.run('networks remove -c 10.20.0.0/17 --force-delete')
        result.exit_code == 1
        assert "cannot forcefully delete a network that does not have a parent" in result.output


##############
# Interfaces #
##############
def test_interfaces_add(site_client, device):
    """Test ``nsot interfaces add``."""
    device_id = device['id']

    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        # Add an interface by id (natural_key not yet supported)
        result = runner.run(
            "interfaces add -D %s -n eth0 -e 'this is eth0'" % device_id
        )
        assert result.exit_code == 0
        assert 'Added interface!' in result.output

        # Verify addition.
        result = runner.run('interfaces list -D %s' % device_id)
        assert result.exit_code == 0
        assert 'eth0' in result.output

        # Create another interface and assign an address to it.
        runner.run('networks add -c 10.10.10.0/24')
        add_result = runner.run(
            'interfaces add -D %s -n eth1 -c 10.10.10.1/32' % device_id
        )
        assert add_result.exit_code == 0

        # Verify addition/assignment.
        result = runner.run('interfaces list -D %s' % device_id)
        assert result.exit_code == 0
        expected = ('eth0', '10.10.10.1/32')
        for e in expected:
            assert e in result.output

        # Create a new interface w/ multiple addresses assigned
        add_result = runner.run(
            'interfaces add -D %s -n eth2 -c 10.10.10.2/32 -c 10.10.10.3/32' %
            device_id
        )
        assert add_result.exit_code == 0

        # Verify it was happy.
        result = runner.run('interfaces list -D %s -n eth2' % device_id)
        assert result.exit_code == 0
        expected = ('10.10.10.2/32', '10.10.10.3/32')
        for e in expected:
            assert e in result.output

        # Test setting parent_id (-p/--parent-id) on create
        parent_ifc = site_client.interfaces.get(name='eth0')[0]
        parent_id = parent_ifc['id']
        result = runner.run(
            'interfaces add -D %s -n eth0:1 -p %s' % (device_id, parent_id)
        )
        assert result.exit_code == 0
        assert 'Added interface!' in result.output


def test_interfaces_list(site_client, device):
    """Test ``nsot interfaces list``."""
    device_id = device['id']
    hostname = device['hostname']

    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        # Add an interface attribute: vlan
        runner.run('attributes add -r interface -n vlan')

        # And a network we can assign addresses from
        runner.run('networks add -c 10.10.10.0/24')

        # Add a couple interfaces to the device
        # eth0: vlan=100, mac=1, speed=10000, type=6 (default)
        i1 = runner.run(
            'interfaces add -D %s -n eth0 -a vlan=100 -m 00:00:00:00:00:01 '
            '-S 10000 -c 10.10.10.1/32' % device_id
        )
        assert i1.exit_code == 0
        # eth1: vlan=100, mac=2, speed=20000, type=24
        i2 = runner.run(
            'interfaces add -D %s -n eth1 -a vlan=100 -m 00:00:00:00:00:02 '
            '-S 20000 -c 10.10.10.2/32 -t 24' % device_id
        )
        assert i2.exit_code == 0

        # Basic list: Make sure both interfaces appear.
        result = runner.run('interfaces list')
        assert result.exit_code == 0
        expected = ('eth0', 'eth1')
        for e in expected:
            assert e in result.output

        ############
        # Querying #
        ############

        # Set query -q/--query
        result = runner.run('interfaces list -q vlan=100')
        expected_output = '{0}:eth0\n{0}:eth1\n'.format(hostname)
        assert result.exit_code == 0
        assert result.output == expected_output

        # Natural key output -N/--natural-key should have same output as -q
        result = runner.run('interfaces list -a vlan=100 -N')
        assert result.exit_code == 0
        assert result.output == expected_output

        # Set query display comma-delimited (-d/--delimited)
        result = runner.run('interfaces list -q vlan=100 -d')
        expected_output = '{0}:eth0,{0}:eth1\n'.format(hostname)
        assert result.exit_code == 0
        assert result.output == expected_output

        # Set query w/ -l/--limit
        result = runner.run('interfaces list -l1 -q vlan=100')
        expected_output = '{0}:eth0\n'.format(hostname)
        assert result.exit_code == 0
        assert result.output == expected_output

        # Set query w/ -l/--limit and -o/--offset
        result = runner.run('interfaces list -l1 -o1 -q vlan=100')
        expected_output = '{0}:eth1\n'.format(hostname)
        assert result.exit_code == 0
        assert result.output == expected_output

        # Grep-friendly output (-g/--grep)
        result = runner.run('interfaces list -a vlan=100 -g')
        expected_output = (
            '{0}:eth0 vlan=100\n'
            '{0}:eth0 addresses=[u\'10.10.10.1/32\']\n'
            '{0}:eth0 description=\n'
            '{0}:eth0 device=16\n'
            '{0}:eth0 device_hostname=foo-bar1\n'
            '{0}:eth0 id=8\n'
            '{0}:eth0 mac_address=00:00:00:00:00:01\n'
            '{0}:eth0 name=eth0\n'
            '{0}:eth0 name_slug=foo-bar1:eth0\n'
            '{0}:eth0 networks=[u\'10.10.10.0/24\']\n'
            '{0}:eth0 parent=None\n'
            '{0}:eth0 parent_id=None\n'
            '{0}:eth0 speed=10000\n'
            '{0}:eth0 type=6\n'
            '{0}:eth1 vlan=100\n'
            '{0}:eth1 addresses=[u\'10.10.10.2/32\']\n'
            '{0}:eth1 description=\n'
            '{0}:eth1 device=16\n'
            '{0}:eth1 device_hostname=foo-bar1\n'
            '{0}:eth1 id=9\n'
            '{0}:eth1 mac_address=00:00:00:00:00:02\n'
            '{0}:eth1 name=eth1\n'
            '{0}:eth1 name_slug=foo-bar1:eth1\n'
            '{0}:eth1 networks=[u\'10.10.10.0/24\']\n'
            '{0}:eth1 parent=None\n'
            '{0}:eth1 parent_id=None\n'
            '{0}:eth1 speed=20000\n'
            '{0}:eth1 type=24\n'
        ).format(hostname)
        assert result.exit_code == 0
        assert result.output == expected_output

        # Query by natural key
        natural_key = '{0}:eth1'.format(hostname)
        result = runner.run('interfaces list -i {}'.format(natural_key))
        assert natural_key in result.output
        assert result.exit_code == 0

        ###########
        # Filtering
        ###########

        # Filter by -D/--device (by id)
        result = runner.run('interfaces list -D %s' % device_id)
        expected = ('eth0', 'eth1')
        assert result.exit_code == 0
        for e in expected:
            assert e in result.output

        # Filter by -D/--device (by hostname) should have same output as by id
        result = runner.run('interfaces list -D %s' % hostname)
        assert result.exit_code == 0
        for e in expected:
            assert e in result.output

        # Filter by -n/--name
        result = runner.run('interfaces list -D %s -n eth1' % hostname)
        assert result.exit_code == 0
        assert 'eth1' in result.output
        assert 'eth0' not in result.output

        # Filter by -S/--speed
        result = runner.run('interfaces list -D %s -S 10000' % hostname)
        assert result.exit_code == 0
        assert 'eth0' in result.output
        assert 'eth1' not in result.output

        # Filter by -t/--type
        result = runner.run('interfaces list -D %s -t 24' % hostname)
        assert result.exit_code == 0
        assert 'eth1' in result.output
        assert 'eth0' not in result.output

        # Filter by -m/--mac-address
        result = runner.run('interfaces list -D %s -m 2' % hostname)
        assert result.exit_code == 0
        assert 'eth1' in result.output
        assert 'eth0' not in result.output


def test_interfaces_subcommands(site_client, device):
    """Test ``nsot interfaces list ... {subcommand}``."""
    device_id = device['id']
    device_hostname = device['hostname']

    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        # Add an interface attribute: vlan
        runner.run('attributes add -r interface -n vlan')

        # And a network for address assignments
        runner.run('networks add -c 10.10.10.0/24')

        # Add a couple interfaces to the device
        # eth0: vlan=100, mac=1, speed=10000, type=6 (default)
        i1 = runner.run(
            'interfaces add -D %s -n eth0 -a vlan=100 -m 00:00:00:00:00:01 '
            '-S 10000 -c 10.10.10.1/32 -c 10.10.10.2/32' % device_id
        )
        assert i1.exit_code == 0

        # Test addresses
        cmds = [
            'interfaces list -D %s -n eth0 -N addresses' % device_id,
            'interfaces list -i %s:eth0 -N addresses' % device_hostname,
            'interfaces list -q vlan=100 -N addresses'
        ]

        for cmd in cmds:
            result = runner.run(cmd)
            assert result.exit_code == 0
            assert result.output == '10.10.10.1/32\n10.10.10.2/32\n'

        # Test networks
        cmds = [
            'interfaces list -D %s -n eth0 -N networks' % device_id,
            'interfaces list -i %s:eth0 -N networks' % device_hostname,
            'interfaces list -q vlan=100 -N networks'
        ]

        for cmd in cmds:
            result = runner.run(cmd)
            assert result.exit_code == 0
            assert result.output == '10.10.10.0/24\n'

        # Test assignments
        cmds = [
            'interfaces list -D %s -n eth0 -N assignments' % device_id,
            'interfaces list -i %s:eth0 -N assignments' % device_hostname,
            'interfaces list -q vlan=100 -N assignments'
        ]
        for cmd in cmds:
            result = runner.run(cmd)
            expected_output = (
                'foo-bar1:eth0:10.10.10.1/32\n'
                'foo-bar1:eth0:10.10.10.2/32\n'
            )
            assert result.exit_code == 0
            assert result.output == expected_output


def test_interfaces_update(site_client, device):
    """Test ``nsot interfaces update``."""
    device_id = device['id']
    hostname = device['hostname']

    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        # Create some attributes
        runner.run('attributes add -n vlan -r interface')
        runner.run('attributes add -n metro -r interface')

        # Create a network for address assignments
        runner.run('networks add -c 10.10.10.0/24')

        # Create an interface w/ attributes set
        # eth0:
        #    vlan=100, metro=lax, mac_address=00:00:00:00:00:01, speed=40000,
        #    type=24, description='this is my eth0', ip=10.10.10.1/32
        runner.run(
            "interfaces add -D %s -n eth0 -a vlan=100 -a metro=lax -m 1 -S "
            "40000 -e 'this is my eth0' -t 24 -c 10.10.10.1/32" % device_id
        )
        parent_ifc = site_client.interfaces.get(name='eth0')[0]
        parent_id = parent_ifc['id']

        # Create a child interface to eth0
        # eth0:1:
        #    ip = 10.10.10.2/32, mac_address=00:00:00:00:00:02
        runner.run(
            "interfaces add -D %s -n eth0:1 -c 10.10.10.2/32" % device_id
        )
        child_ifc = site_client.interfaces.get(name='eth0:1')[0]
        child_id = child_ifc['id']

        # Test attributes: update vlan=N
        cases = [
            [200, parent_id],
            [300, '%s:%s' % (hostname, parent_ifc['name'])],
        ]

        for vlan, identifier in cases:
            result = runner.run(
                'interfaces update -i %s -a vlan=%d' % (identifier, vlan)
            )
            assert result.exit_code == 0
            assert 'Updated interface!' in result.output

            # Verify attribute update
            result = runner.run('interfaces list -i %s' % identifier)
            assert result.exit_code == 0
            assert 'vlan=%d' % (vlan) in result.output

        # Test parent: update eth0:1 parent to eth0
        result = runner.run(
            'interfaces update -i %s -p %s' % (child_id, parent_id)
        )
        assert result.exit_code == 0

        # Verify parent: eth0:1 parent should be eth0
        result = runner.run('interfaces list -p %s' % parent_id)
        assert result.exit_code == 0
        assert 'eth0:1' in result.output

        # Update name, mac_address, type, speed
        result = runner.run(
            "interfaces update -i %s -n child -m 3 -t 161 -S 12345678" %
            child_id
        )
        assert result.exit_code == 0

        # Verify name and mac_address updated
        result = runner.run('interfaces list -n child')  # Lookup by new name
        assert result.exit_code == 0
        expected = (
            '161',  # type
            '12345678',  # speed
            '00:00:00:00:00:03'  # mac_address
        )
        for e in expected:
            assert e in result.output

        # Test addresses  - We know they will be empty
        # FIXME(jathan): Once we have a better story about differential
        # assignment of addresses to interfaces, make it so that addresses can
        # be persistent on updates.
        result = runner.run('interfaces list -i %s' % parent_id)
        assert result.exit_code == 0
        assert '10.10.10.1/32' not in result.output

        # So let's add it back and verify...
        runner.run(
            'interfaces update -i %s -c 10.10.10.1/32' % parent_id
        )
        result = runner.run('interfaces list -i %s' % parent_id)
        assert result.exit_code == 0
        assert '10.10.10.1/32' in result.output

        # Test description.
        # FIXME(jathan): It doesn't currently show in the CLI output. So we're
        # just making sure it doesn't fail.
        result = runner.run(
            "interfaces update -i %s -e 'description'" % child_id
        )
        assert result.exit_code == 0


def test_interfaces_remove(site_client, device, interface):
    """Test ``nsot interfaces remove``."""
    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        # Just delete the interface we have.
        result = runner.run('interfaces remove -i %s' % interface['id'])
        assert result.exit_code == 0
        assert 'Removed interface!' in result.output


def test_interfaces_remove_by_natural_key(site_client, device, interface):
    """Test ``nsot interfaces remove`` via the natural key."""
    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        # Just delete the interface we have, but by natural key this time.
        identifier = '%s:%s' % (device['hostname'], interface['name'])
        result = runner.run('interfaces remove -i %s' % identifier)
        assert result.exit_code == 0
        assert 'Removed interface!' in result.output


##########
# Values #
##########
def test_values_list(site_client):
    """Test ``nsot values list``."""
    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        # Create the owner attribute
        runner.run('attributes add -n owner -r device')

        # Create a single device w/ owner= set
        runner.run('devices add -H foo-bar1 -a owner=jathan')

        # Make sure -n/--name is required.
        result = runner.run('values list')
        assert result.exit_code == 2
        assert 'Error: Missing option "-n"' in result.output

        # Run a simple list to get the expected result.
        result = runner.run('values list -n owner -r device')
        assert result.exit_code == 0
        assert result.output == 'jathan\n'


###########
# Changes #
###########
def test_changes_list(site_client):
    """Test ``nsot changes list``."""
    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        # Just make sure it works.
        result = runner.run('changes list')
        assert result.exit_code == 0



# -*- coding: utf-8 -*-

"""
Make dummy data and fixtures and stuff.
"""

from __future__ import unicode_literals
import logging
import os

import pytest
from pytest_django.fixtures import live_server, django_user_model

from pynsot.client import get_api_client
from tests.util import CliRunner

__all__ = ('django_user_model', 'live_server')


# Logger
log = logging.getLogger(__name__)

# API version to use for the API client
API_VERSION = os.getenv('NSOT_API_VERSION')

# Dummy config data used for testing auth_header authentication.
AUTH_HEADER_CONFIG = {
    'auth_method': 'auth_header',
    'default_domain': 'localhost',
    'auth_header': 'X-NSoT-Email',
}

# This is used to test dotfile settings.
DOTFILE_CONFIG_DATA = {
    'auth_token': {
        'email': 'jathan@localhost',
        'url': 'http://localhost:8990/api',
        'auth_method': 'auth_token',
        'secret_key': 'MJMOl9W7jqQK3h-quiUR-cSUeuyDRhbn2ca5E31sH_I=',
    },
    'auth_header': {
        'email': 'jathan@localhost',
        'url': 'http://localhost:8990/api',
        'auth_method': 'auth_header',
        'default_domain': 'localhost',
        'auth_header': 'X-NSoT-Email',
    }
}


@pytest.fixture
def config(live_server, django_user_model):
    """Create a user and return an auth_token config matching that user."""
    user = django_user_model.objects.create(
        email='jathan@localhost', is_superuser=True, is_staff=True
    )
    data = {
        'email': user.email,
        'secret_key': user.secret_key,
        'auth_method': 'auth_token',
        'url': live_server.url + '/api',
        # 'api_version': API_VERSION,
        'api_version': '1.0',  # Hard-coded.
    }

    return data


@pytest.fixture
def auth_header_config(config):
    """Return an auth_header config."""
    config.pop('secret_key')
    config.update(AUTH_HEADER_CONFIG)
    return config


@pytest.fixture
def client(config):
    """Create and return an admin client."""
    api = get_api_client(extra_args=config, use_dotfile=False)
    api.config = config
    return api


@pytest.fixture
def site(client):
    """Returns a Site object."""
    return client.sites.post({'name': 'Foo', 'description': 'Foo site.'})


@pytest.fixture
def site_client(client, site):
    """Returns a client tied to a specific site."""
    client.config['default_site'] = site['id']
    client.default_site = site['id']
    return client


@pytest.fixture
def runner(site_client):
    return CliRunner(site_client.config)


@pytest.fixture
def attribute(site_client):
    """Return an Attribute object."""
    return site_client.sites(site_client.default_site).attributes.post(
        {'name': 'foo', 'resource_name': 'Device'}
    )


@pytest.fixture
def attributes(site_client):
    """ A bunch of attributes for each resource type """

    results = []
    resources = (
        'Circuit',
        'Device',
        'Interface',
        'Network',
        'Protocol',
    )

    for r in resources:
        attr = site_client.sites(site_client.default_site).attributes.post(
            {'name': 'foo', 'resource_name': r}
        )

        results.append(attr)

    return results


@pytest.fixture
def device(site_client, attributes):
    """Return a Device object."""
    return site_client.sites(site_client.default_site).devices.post(
        {
            'hostname': 'foo-bar1',
            'attributes': {
                'foo': 'test_device'
            },
        }
    )


@pytest.fixture
def network(site_client, attributes):
    """Return a Network object."""
    return site_client.sites(site_client.default_site).networks.post(
        {
            'cidr': '10.20.30.0/24',
            'attributes': {
                'foo': 'test_network'
            }
        }
    )


@pytest.fixture
def interface(site_client, attributes, device, network):
    """
    Return an Interface object.

    Interface is bound to ``device`` with an address assigned from ``network``.
    """
    device_id = device['id']
    return site_client.sites(site_client.default_site).interfaces.post(
        {
            'name': 'eth0',
            'addresses': ['10.20.30.1/32'],
            'device': device_id,
            'attributes': {'foo': 'test_interface'},
        }
    )


@pytest.fixture
def protocol_type(site_client):
    """
    Return a Protocol Type Object.
    """
    return site_client.sites(site_client.default_site).protocol_types.post(
        {
            'name': 'bgp',
        }
    )



# -*- coding: utf-8 -*-

"""
Test Circuits in the CLI app.
"""

from __future__ import absolute_import, unicode_literals
import logging

import pytest

from tests.fixtures import (attribute, attributes, client, config, device,
                            interface, network, site, site_client)

from tests.fixtures.circuits import circuit

from tests.fixtures.protocol_types import (protocol_type, protocol_types,
                                           protocol_attribute,
                                           protocol_attribute2)

from tests.util import CliRunner, assert_output

log = logging.getLogger(__name__)


def test_protocol_types_add(site_client, protocol_attribute):
    """Test ``nsot protocol_types add``."""

    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        # Add a protocol_type by name.
        result = runner.run("protocol_types add -n bgp")
        assert result.exit_code == 0
        assert 'Added protocol_type!' in result.output

        # Verify addition.
        result = runner.run('protocol_types list')
        assert result.exit_code == 0
        assert 'bgp' in result.output

        # Add a protocol with same name and fail.
        result = runner.run("protocol_types add -n bgp")
        expected_output = 'The fields site, name must make a unique set.'
        assert result.exit_code != 0
        assert expected_output in result.output

        # Add second protocol_type by name.
        result = runner.run("protocol_types add -n ospf -d 'OSPF is the best'")
        assert result.exit_code == 0
        assert 'Added protocol_type!' in result.output

        # Verify default site is assigned and verify description.
        site_id = str(protocol_attribute['site_id'])
        result = runner.run('protocol_types list -i 2')
        assert result.exit_code == 0
        assert site_id in result.output
        assert 'OSPF is the best' in result.output

        # Add a third protocol type with required_attribute.
        result = runner.run('protocol_types add -n tcp -r %s' % protocol_attribute['name'])
        assert result.exit_code == 0
        assert 'Added protocol_type!' in result.output

        # Verify protocol attribute is assigned.
        result = runner.run('protocol_types list -n tcp')
        assert result.exit_code == 0
        assert protocol_attribute['name'] in result.output


def test_protocol_types_list(site_client, protocol_type, protocol_attribute, protocol_attribute2):
    """Test ``nsot protocol_types list``"""

    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        # Basic List.
        result = runner.run('protocol_types list')
        assert result.exit_code == 0
        assert protocol_type['name'] in result.output

        # Test -d/--description
        result = runner.run('protocol_types list -d "%s"' % protocol_type['description'])
        assert result.exit_code == 0
        assert protocol_type['description'] in result.output

        # Test -n/--name
        result = runner.run('protocol_types list -n %s' % protocol_type['name'])
        assert result.exit_code == 0
        assert protocol_type['name'] in result.output

        # Test -s/--site
        result = runner.run('protocol_types list -s %s' % protocol_type['site'])
        assert result.exit_code == 0
        assert protocol_type['name'] in result.output

        # Test -i/--id
        result = runner.run('protocol_types list -i %s' % protocol_type['id'])
        assert result.exit_code == 0
        assert protocol_type['name'] in result.output


def test_protocol_types_list_limit(site_client, protocol_types):
    """
    If ``--limit 2`` is used, we should only see the first two Protocol objects
    """
    limit = 2
    runner = CliRunner(site_client.config)

    with runner.isolated_filesystem():
        result = runner.run('protocol_types list -l {}'.format(limit))
        assert result.exit_code == 0

        expected_types = protocol_types[:limit]
        unexpected_types = protocol_types[limit:]

        for t in expected_types:
            assert t['name'] in result.output
        for t in unexpected_types:
            assert t['name'] not in result.output


def test_protocol_types_list_offset(site_client, protocol_types):
    """
    If ``--limit 2`` and ``--offset 2`` are used, we should only see the third
    and fourth Protocol objects that were created
    """
    limit = 2
    offset = 2
    runner = CliRunner(site_client.config)

    with runner.isolated_filesystem():
        result = runner.run(
            'protocol_types list -l {} -o {}'.format(limit, offset)
        )
        assert result.exit_code == 0

        expected_types = protocol_types[offset:limit+offset]
        unexpected_types = protocol_types[limit+offset:]

        for t in expected_types:
            assert t['name'] in result.output
        for t in unexpected_types:
            assert t['name'] not in result.output


def test_protocol_types_update(site_client, protocol_type, protocol_attribute):
    """Test ``nsot protocol_types update``"""

    pt_id = protocol_type['id']

    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        # Try to change the name
        result = runner.run('protocol_types update -n Cake -i %s' % (pt_id))
        assert result.exit_code == 0
        assert 'Updated protocol_type!' in result.output

        # Update the description
        result = runner.run('protocol_types update -d Rise -i %s' % (pt_id))
        assert result.exit_code == 0
        assert 'Updated protocol_type!' in result.output

        # Assert the Cake Rises
        result = runner.run('protocol_types list -i %s' % pt_id)
        assert result.exit_code == 0
        assert 'Cake' in result.output
        assert 'Rise' in result.output

        # Test add attributes
        result = runner.run('protocol_types update -r %s -i %s' % (
                protocol_attribute ['name'],
                pt_id,
            )
        )
        assert result.exit_code == 0
        assert 'Updated protocol_type!' in result.output

        # Assert the attribute was added
        result = runner.run('protocol_types list -i %s' % pt_id)
        assert result.exit_code == 0
        assert protocol_attribute['name'] in result.output


def test_protocol_types_remove(site_client, protocol_type):
    """Test ``nsot protocol_types remove``"""

    runner = CliRunner(site_client.config)
    with runner.isolated_filesystem():
        result = runner.run(
            'protocol_types remove -i %s' % (protocol_type['id'])
        )
        assert result.exit_code == 0
        assert 'Removed protocol_type!' in result.output



# -*- coding: utf-8 -*-

from __future__ import absolute_import, unicode_literals

import pytest

from tests.fixtures import site_client

@pytest.fixture
def protocol_type(site_client, protocol_attribute, protocol_attribute2):
    """
    Return a Protocol Type Object.
    """
    return site_client.sites(site_client.default_site).protocol_types.post(
        {
            'name': 'bgp',
            'required-attributes': ['boo', 'foo',],
            'description': 'bgp is my bestie',
        }
    )


@pytest.fixture
def protocol_attribute(site_client):
    return site_client.sites(site_client.default_site).attributes.post(
        {
            'name':'boo',
            'value': 'test_attribute',
            'resource_name': 'Protocol',
        }
    )

@pytest.fixture
def protocol_attribute2(site_client):
    return site_client.sites(site_client.default_site).attributes.post(
        {
            'name':'foo',
            'value': 'test_attribute',
            'resource_name': 'Protocol',
        }
    )


@pytest.fixture
def protocol_types(site_client):
    """
    A group of ProtocolTypes for testing limit/offset/etc.
    """
    protocol_types = []
    site = site_client.sites(site_client.default_site)

    for i in range(1, 6):
        protocol_type = site.protocol_types.post({
            'name': 'type{}'.format(i),
        })
        protocol_types.append(protocol_type)

    return protocol_types



# -*- coding: utf-8 -*-

from __future__ import absolute_import, unicode_literals

import pytest

from tests.fixtures import site_client


@pytest.fixture
def circuit_attributes(site_client):
    attr_names = [
        'owner',
        'vendor'
    ]

    attrs = ({'name': a, 'resource_name': 'Circuit'} for a in attr_names)
    client = site_client.sites(site_client.default_site).attributes
    return map(client.post, attrs)


@pytest.fixture
def device_a(site_client):
    """ Device for the A side of the circuit """

    return site_client.sites(site_client.default_site).devices.post(
        {'hostname': 'foo-bar01'}
    )


@pytest.fixture
def device_z(site_client):
    """ Device for the Z side of the circuit """

    return site_client.sites(site_client.default_site).devices.post(
        {'hostname': 'foo-bar02'}
    )


@pytest.fixture
def interface_a(site_client, device_a, network):
    """ Interface for the A side of the circuit """

    return site_client.sites(site_client.default_site).interfaces.post({
        'device': device_a['id'],
        'name': 'eth0',
        'addresses': ['10.20.30.2/32'],
    })


@pytest.fixture
def interface_z(site_client, device_z, network):
    """ Interface for the Z side of the circuit """

    return site_client.sites(site_client.default_site).interfaces.post({
        'device': device_z['id'],
        'name': 'eth0',
        'addresses': ['10.20.30.3/32'],
    })


@pytest.fixture
def circuit(site_client, circuit_attributes, interface_a, interface_z):
    """ Circuit connecting interface_a to interface_z """

    return site_client.sites(site_client.default_site).circuits.post({
        'name': 'test_circuit',
        'endpoint_a': interface_a['id'],
        'endpoint_z': interface_z['id'],
        'attributes': {
            'owner': 'alice',
            'vendor': 'lasers go pew pew',
        },
    })


@pytest.fixture
def dangling_circuit(site_client, interface):
    """
    Circuit where we only own the local site, remote (Z) end is a vendor
    """

    return site_client.sites(site_client.default_site).circuits.post({
        'name': 'remote_vendor_circuit',
        'endpoint_a': interface['id'],
    })


@pytest.fixture
def attributeless_circuit(site_client, interface):
    """ Circuit with no attributes set """

    return site_client.sites(site_client.default_site).circuits.post({
        'name': 'attributeless_circuit',
        'endpoint_a': interface['id'],
        'attributes': {},
    })
